import copy
import logging

import numpy as np

from seaduck.runtime_conf import compileable

# legal_tends = [0, 1, 2, 3]  # up,down,left,right #list(LLC_FACE_CONNECT.columns)

LLC_FACE_CONNECT = np.array(
    [
        [1, 42, 12, 3],
        [2, 0, 11, 4],
        [6, 1, 10, 5],
        [4, 42, 0, 9],
        [5, 3, 1, 8],
        [6, 4, 2, 7],
        [10, 5, 2, 7],
        [10, 5, 6, 8],
        [11, 4, 7, 9],
        [12, 3, 8, 42],
        [2, 7, 6, 11],
        [1, 8, 10, 12],
        [0, 9, 11, 42],
    ]
)

DIRECTIONS = np.array([np.pi / 2, -np.pi / 2, np.pi, 0])


@compileable
def _llc_mutual_direction(face, neighbor_face, transitive=False):
    """Find the relative orientation of two faces.

    The compileable version of mutual direction for llc grid.
    See Topology.mutual direction for more detail.
    """
    connected_to_face = np.where(LLC_FACE_CONNECT[face] == neighbor_face)
    connected_to_neigh = np.where(LLC_FACE_CONNECT[neighbor_face] == face)
    if len(connected_to_face[0]) == 0:
        found = False
    else:
        found = connected_to_face[0][0] in [0, 1, 2, 3] and connected_to_neigh[0][
            0
        ] in [0, 1, 2, 3]
    if found:
        return connected_to_face[0][0], connected_to_neigh[0][0]
    elif transitive:
        common = -1
        for i in LLC_FACE_CONNECT[face]:
            if i in LLC_FACE_CONNECT[neighbor_face]:
                common = i
                break
        if common < 0:
            raise ValueError(
                "The two faces does not share common face, transitive did not help"
            )
        else:
            edge_1 = np.where(LLC_FACE_CONNECT[face] == common)[0][0]
            new_edge_1 = np.where(LLC_FACE_CONNECT[common] == face)[0][0]
            edge_2 = np.where(LLC_FACE_CONNECT[common] == neighbor_face)[0][0]
            new_edge_2 = np.where(LLC_FACE_CONNECT[neighbor_face] == common)[0][0]
            if (edge_1 in [0, 1] and new_edge_1 in [0, 1]) or (
                edge_1 in [2, 3] and new_edge_1 in [2, 3]
            ):
                return edge_2, new_edge_2
            elif (edge_2 in [0, 1] and new_edge_2 in [0, 1]) or (
                edge_2 in [2, 3] and new_edge_2 in [2, 3]
            ):
                return edge_1, new_edge_1
            else:
                raise NotImplementedError(
                    "the common face is not parallel to either of the face"
                )
    else:
        raise ValueError("The two faces are not connected (transitive = False)")


@compileable
def _llc_get_the_other_edge(face, edge):
    """See what is adjacent to the face by this edge.

    The compileable version of get_the_other_edge for llc grid.
    See Topology.get_the_other_edge for more detail.
    """
    face_connect = LLC_FACE_CONNECT
    neighbor_face = face_connect[face, edge]
    if neighbor_face == 42:
        raise IndexError(
            "Reaching the edge where the face is not connected to any other face"
        )
    connected_to_neigh = np.where(face_connect[neighbor_face] == face)
    return neighbor_face, connected_to_neigh[0][0]


@compileable
def _box_ind_tend(ind, tend, iymax, ixmax):
    """Move an index in a direction.

    The compileable version of ind_tend for regional (box) grid.
    See Topology.ind_tend for more detail.
    """
    iy, ix = ind
    if tend == 0:
        iy += 1
    elif tend == 1:
        iy -= 1
    elif tend == 2:
        ix -= 1
    elif tend == 3:
        ix += 1
    # it would be better to raise an error here.
    if (iy > iymax) or (iy < 0):
        return (-1, -1)
    if (ix > ixmax) or (ix < 0):
        return (-1, -1)
    return (iy, ix)


@compileable
def _x_per_ind_tend(ind, tend, iymax, ixmax):
    """Move an index in a direction.

    The compileable version of ind_tend for zonally periodic (x-per) grid.
    See Topology.ind_tend for more detail.
    """
    iy, ix = ind
    if tend == 0:
        iy += 1
    elif tend == 1:
        iy -= 1
    elif tend == 2:
        ix -= 1
    elif tend == 3:
        ix += 1
    if (iy > iymax) or (iy < 0):
        return (-1, -1)
    if ix > ixmax:
        return (iy, ix - ixmax)
    if ix < 0:
        return (iy, ixmax + ix + 1)
    return (iy, ix)


@compileable
def _llc_ind_tend(ind, tendency, iymax, ixmax):
    """Move an index in a direction.

    The compileable version of ind_tend for llc grid.
    See Topology.ind_tend for more detail.
    """
    #     iymax = 89
    #     ixmax = 89
    face, iy, ix = ind
    if tendency == 3:
        if ix != ixmax:
            ix += 1
        else:
            neighbor_face, new_edge = _llc_get_the_other_edge(face, 3)
            if new_edge == 1:
                face, iy, ix = [neighbor_face, 0, ixmax - iy]
            elif new_edge == 0:
                face, iy, ix = [neighbor_face, iymax, iy]
            elif new_edge == 2:
                face, iy, ix = [neighbor_face, iy, 0]
            elif new_edge == 3:
                face, iy, ix = [neighbor_face, iymax - iy, ixmax]
    if tendency == 2:
        if ix != 0:
            ix -= 1
        else:
            neighbor_face, new_edge = _llc_get_the_other_edge(face, 2)
            if new_edge == 1:
                face, iy, ix = [neighbor_face, 0, iy]
            elif new_edge == 0:
                face, iy, ix = [neighbor_face, iymax, ixmax - iy]
            elif new_edge == 2:
                face, iy, ix = [neighbor_face, iymax - iy, 0]
            elif new_edge == 3:
                face, iy, ix = [neighbor_face, iy, ixmax]
    if tendency == 0:
        if iy != iymax:
            iy += 1
        else:
            neighbor_face, new_edge = _llc_get_the_other_edge(face, 0)
            if new_edge == 1:
                face, iy, ix = [neighbor_face, 0, ix]
            elif new_edge == 0:
                face, iy, ix = [neighbor_face, iymax, ixmax - ix]
            elif new_edge == 2:
                face, iy, ix = [neighbor_face, iymax - ix, 0]
            elif new_edge == 3:
                face, iy, ix = [neighbor_face, ix, ixmax]
    if tendency == 1:
        if iy != 0:
            iy -= 1
        else:
            neighbor_face, new_edge = _llc_get_the_other_edge(face, 1)
            if new_edge == 1:
                face, iy, ix = [neighbor_face, 0, ixmax - ix]
            elif new_edge == 0:
                face, iy, ix = [neighbor_face, iymax, ix]
            elif new_edge == 2:
                face, iy, ix = [neighbor_face, ix, 0]
            elif new_edge == 3:
                face, iy, ix = [neighbor_face, iymax - ix, ixmax]
    return (face, iy, ix)


@compileable
def _llc_get_uv_mask_from_face(faces):
    """Get the masking of UV points.

    The compileable version of get_uv_mask_from_face for llc grid.
    See Topology.get_uv_mask_from_face for more detail.
    """
    # we are considering a row from the fatten_face
    # faces is essentially which face each node of the kernel is on.
    n = len(faces)  # it should have been m to be more consistent with other code
    UfromUvel = np.ones(n)
    UfromVvel = np.zeros(n)
    VfromUvel = np.zeros(n)
    VfromVvel = np.ones(n)
    # if all of the node are on the same face, we don't have to convert anything
    if np.abs(np.ones(n) * faces[0] - faces).max() < 1e-5:
        return UfromUvel, UfromVvel, VfromUvel, VfromVvel
    else:
        for i in range(1, n):
            if faces[i] == faces[0]:
                continue
            # if the face is not the same, we need to do something
            else:
                # get how much the new face is rotated from the old face
                edge, new_edge = _llc_mutual_direction(
                    faces[0], faces[i], transitive=True
                )
                rot = np.pi - DIRECTIONS[edge] + DIRECTIONS[new_edge]
                # you can think of this as a rotation matrix
                UfromUvel[i] = np.cos(rot)
                UfromVvel[i] = np.sin(rot)
                VfromUvel[i] = -np.sin(rot)
                VfromVvel[i] = np.cos(rot)
        return UfromUvel, UfromVvel, VfromUvel, VfromVvel


class Topology:
    """Topology object.

    A light weight object that remembers the structure of the grid,
    what is connected, what is not. The core method is simply move the index to a direction.
    In the movie kill Bill, the bride sat in a car and said "wiggle your big toe".
    After the toe moved, "the hard part is over". You get the idea.

    Parameters
    ----------
    od: xarray.Dataset
        The dataset to record topological info from.
    typ: None, or str
        Type of the grid.
        Currently we support
        'box' for regional dataset,
        'x-periodic' for zonally periodic ones,
        'llc' for lat-lon-cap dataset.
        We recommend that users put None here,
        so that the type is figured out automatically.
    """

    def __init__(self, od, typ=None):
        try:
            h_shape = od["XC"].shape
        except KeyError:
            try:
                h_shape = (int(od["lat"].shape[0]), int(od["lon"].shape[0]))
            except KeyError as exc:
                raise KeyError(
                    "Either XC or lat/lon is needed to create the Topology object"
                ) from exc
        self.h_shape = h_shape
        if "XG" in od.variables:
            self.g_shape = od["XG"].shape
        else:
            self.g_shape = self.h_shape
        try:
            self.itmax = len(od["time"]) - 1
        except (KeyError, TypeError):
            self.itmax = 0
        try:
            self.izmax = len(od["Z"]) - 1
        except (KeyError, TypeError):
            self.izmax = 0

        if typ:
            self.typ = typ
        elif typ is None:
            if len(h_shape) == 3:
                self.num_face, self.iymax, self.ixmax = h_shape
                self.iymax -= 1
                self.ixmax -= 1
                if self.num_face == 13:
                    self.typ = "LLC"
                    # we can potentially generate the face connection in runtime
                    # say, put the csv file on cloud
                else:
                    raise NotImplementedError(
                        "Other complex topology is not implemented yet."
                        "If you want to work with such data, please contact the authors."
                    )
            elif len(h_shape) == 2:
                self.iymax, self.ixmax = h_shape
                self.iymax -= 1
                self.ixmax -= 1
                try:
                    lon_right = float(od["XC"][0, self.ixmax])
                    lon_left = float(od["XC"][0, 0])
                except KeyError:
                    lon_right = float(od["lon"][self.ixmax])
                    lon_left = float(od["lon"][0])
                left_to_right = (lon_right - lon_left) % 360
                right_to_left = (lon_left - lon_right) % 360
                if left_to_right > 50 * right_to_left:
                    self.typ = "x_periodic"
                else:
                    self.typ = "box"

    def get_the_other_edge(self, face, edge):
        """See what is adjacent to the face by this edge.

        The (edge) side of the (face) is connected to
        the (new_edge) side of the (new_face).
        0,1,2,3 stands for up, down, left, right.

        Parameters
        ----------
        face: int
            The face of interst
        edge: 0,1,2,3
            which direction of the face we are looking for

        Returns
        -------
        new_face: int
            The face adjacent to face in the edge direction.
        new_edge: 0,1,2,3
            The face is connected to new_face in which direction.
        """
        if self.typ == "LLC":
            return _llc_get_the_other_edge(face, edge)
        elif self.typ in ["x_periodic", "box"]:
            raise ValueError(
                "It makes no sense to tinker with face_connection when there is only one face"
            )
        else:
            raise NotImplementedError

    def mutual_direction(self, face, new_face, **kwarg):
        """Find the relative orientation of two faces.

        0,1,2,3 stands for up, down, left, right
        given 2 faces, the returns are
        1. the 2nd face is to which direction of the 1st face
        2. the 1st face is to which direction of the 2nd face.
        """
        if self.typ == "LLC":
            return _llc_mutual_direction(face, new_face, **kwarg)
        elif self.typ in ["x_periodic", "box"]:
            raise ValueError(
                "It makes no sense to tinker with face_connection when there is only one face"
            )
        else:
            raise NotImplementedError

    def ind_tend(self, ind, tend, cuvwg="C", **kwarg):
        """Move an index in a direction.

        ind is a tuple that is face,iy,ix,
        tendency again is up, down, left, right represented by 0,1,2,3
        return the next cell.

        Parameters
        ----------
        ind: tuple
            The index to find the neighbor of
        tend: int
            Which direction to move from the original index.
        cuvwg: str, default "C"
            Whether we are moving from C grid, U grid, V grid, or G grid.
        kwarg:dict, optional
            Keyword argument that currently only apply for the llc case.
        """
        if -1 in ind:
            # meaning invalid point
            return tuple(-1 for i in ind)
        if self.typ == "LLC":
            if cuvwg == "C":
                to_return = _llc_ind_tend(ind, tend, self.iymax, self.ixmax, **kwarg)
            elif cuvwg == "U":
                _, to_return = self._ind_tend_U(ind, tend)
            elif cuvwg == "V":
                _, to_return = self._ind_tend_V(ind, tend)
            elif cuvwg == "G":
                to_return = self._ind_tend_G(ind, tend)
            else:
                raise ValueError("The type of grid point should be among C,U,V,G")
        elif self.typ == "x_periodic":
            to_return = _x_per_ind_tend(ind, tend, self.iymax, self.ixmax, **kwarg)
        elif self.typ == "box":
            to_return = _box_ind_tend(ind, tend, self.iymax, self.ixmax, **kwarg)
        else:
            raise NotImplementedError
        return tuple(int(i) for i in to_return)

    def ind_moves(self, ind, moves, **kwarg):
        """Move an index in a serie of directions.

        moves being a list of directions (0,1,2,3),
        ind being the starting index,
        return the index after moving in the directions in the list.

        Parameters
        ----------
        ind: tuple
            Index of the starting position
        moves: iterable
            A sequence of steps to "walk" from the original position.
        kwarg: dict, optional
            Keyword arguments that pass into ind_tend.
        """
        if self.check_illegal(ind):
            return tuple(-1 for i in ind)  # the origin is invalid
        if not set(moves).issubset({0, 1, 2, 3}):
            raise ValueError("Illegal move. Must be 0,1,2,3")
        if self.typ in ["LLC", "cubed_sphere"]:
            face, iy, ix = copy.deepcopy(ind)
            movess = copy.deepcopy(moves)
            for k in range(len(movess)):
                move = movess[k]
                ind = self.ind_tend(ind, move, **kwarg)
                if ind[0] != face:  # if the face has changed
                    # there are times where the the kernel lies between
                    # 2 faces that define 'left' differently. That's why
                    # when that happens we need to correct the direction
                    # you want to move the indexes.
                    edge, new_edge = self.mutual_direction(
                        face, ind[0], transitive=True
                    )
                    rot = (np.pi - DIRECTIONS[edge] + DIRECTIONS[new_edge]) % (
                        np.pi * 2
                    )
                    if np.isclose(rot, 0):
                        pass
                    elif np.isclose(rot, np.pi / 2):
                        movess[k + 1 :] = [
                            [2, 3, 1, 0][move] for move in movess[k + 1 :]
                        ]
                    elif np.isclose(rot, 3 * np.pi / 2):
                        movess[k + 1 :] = [
                            [3, 2, 0, 1][move] for move in movess[k + 1 :]
                        ]
                    face = ind[0]
                    # if the old face is on the left of the new face,
                    # the particle should be heading right
        elif self.typ in ["x_periodic", "box"]:
            for move in moves:
                ind = self.ind_tend(ind, move)
        return tuple(int(i) for i in ind)

    def check_illegal(self, ind, cuvwg="C"):
        """Check if the index is legal.

        A vectorized check to see whether the index is legal,
        index can be a tuple of numpy ndarrays.
        no negative index is permitted for sanity reason.

        Parameters
        ----------
        ind: tuple
            Each element of the tuple is iterable of one dimension of the indexes.
        cuvwg: 'C' or 'G'
            Whether use the center grid or the corner grid.
        """
        if cuvwg == "C":
            the_shape = self.h_shape
        else:
            the_shape = self.g_shape

        if isinstance(ind[0], int):  # for single item
            result = False
            for i, z in enumerate(ind):
                max_pos = the_shape[i]
                if not 0 <= z <= max_pos - 1:
                    result = True
            return result
        else:  # for numpy ndarray
            result = False
            for i, z in enumerate(ind):
                max_pos = the_shape[i]
                result = np.logical_or(
                    np.logical_or((0 > z), (z > max_pos - 1)), result
                )
            return result

    def ind_tend_vec(self, inds, tend, **kwarg):
        """Move many indices in a list of directions.

        Vectorized version for ind_tend method.

        Parameters
        ----------
        inds: tuple of numpy.array or numpy.array
            Each element of the tuple is iterable of one dimension of the indexes.
        tend: iterable
            Which direction should each index take.
        kwarg: dict,optional
            Keyword argument that feeds into ind_tend.
        """
        inds = np.array(inds)
        old_inds = copy.deepcopy(inds)
        move_dic = {
            0: np.array([1, 0]),  # delta_y,delta_x
            1: np.array([-1, 0]),
            2: np.array([0, -1]),
            3: np.array([0, 1]),
        }
        naive_move = np.array([move_dic[i] for i in tend], dtype=int).T
        inds[-2:] += naive_move
        cuvwg = kwarg.get("cuvwg", "C")
        illegal = self.check_illegal(inds, cuvwg=cuvwg)
        redo = np.array(np.where(illegal)).T
        particle_on_edge = False
        for num, loc in enumerate(redo):
            j = loc[0]
            ind = tuple(old_inds[:, j])
            try:
                n_ind = self.ind_tend(ind, int(tend[j]), **kwarg)
            except IndexError:
                particle_on_edge = True
                n_ind = ind
            inds[:, j] = np.array(n_ind).ravel()
        if particle_on_edge:
            logging.warning("Some points are on the edge")
        for i in range(len(inds)):
            inds[i] = inds[i].astype(int)
        return inds

    def _find_wall_between(self, ind1, ind2):
        """Return the wall between two adjacent cells.

        if the input is not adjacent, error may not be raised
        This scheme is only valid if there is face in the dimensions.
        """
        (fc1, _, _) = ind1
        (fc2, _, _) = ind2
        non_normal_connection = ValueError(
            f"The two face connecting the indexes {ind1},{ind2}"
            " are not connected in a normal way"
        )
        if fc1 == fc2:
            to_return = tuple(
                np.ceil((np.array(ind1) + np.array(ind2)) / 2).astype(int)
            )
            other = ind1 if ind2 == to_return else ind2
            (_, iyr, ixr) = to_return
            (_, iyo, ixo) = other
            if ixr > ixo:
                return "U", to_return
            elif iyr > iyo:
                return "V", to_return
            else:
                raise IndexError(
                    f"there is no wall between a cell and itself,{ind1},{ind2}"
                )
                # This error can be raised when there is three instead of four points in a corner
        else:
            d1to2, d2to1 = self.mutual_direction(fc1, fc2)
            if d1to2 in [0, 3]:
                to_return = ind2
                if d2to1 == 1:
                    return "V", to_return
                elif d2to1 == 2:
                    return "U", to_return
                else:
                    raise non_normal_connection
            elif d2to1 in [0, 3]:
                to_return = ind1
                if d1to2 == 1:
                    return "V", to_return
                elif d1to2 == 2:
                    return "U", to_return
                else:
                    raise non_normal_connection
            else:
                raise non_normal_connection

    def _ind_tend_U(self, ind, tend):
        """Move an U-index in a direction.

        A subset of ind_tend that deal with special issues
        that comes from staggered grid. Read ind_tend for more info.
        """
        # TODO: implement different grid offset case
        first = self.ind_tend(ind, tend)
        if first[0] == ind[0]:
            return "U", first
        else:
            alter = self.ind_moves(ind, [2, tend])
            if alter == first:
                return "U", alter
            else:
                return self._find_wall_between(first, alter)

    def _ind_tend_V(self, ind, tend):
        """Move a V-index in a direction.

        A subset of ind_tend that deal with special issues
        that comes from staggered grid. Read ind_tend for more info.
        """
        # TODO: implement different grid offset case
        first = self.ind_tend(ind, tend)
        if first[0] == ind[0]:
            return "V", first
        else:
            alter = self.ind_moves(ind, [1, tend])
            return self._find_wall_between(first, alter)

    def _ind_tend_G(self, ind, tend):
        """Move a G-index(corner point) in a direction.

        If there is a index for the corner point at all,
        it needs to be connected to the two edges with the same index.
        Therefore, this will always work, regardless of the grid.
        """
        if tend in [0, 1]:
            _, nind = self._ind_tend_U(ind, tend)
            return nind
        elif tend in [2, 3]:
            _, nind = self._ind_tend_V(ind, tend)
            return nind
        else:
            raise ValueError(f"tend {tend} not supported")

    def get_uv_mask_from_face(self, faces):
        """Get the masking of UV points.

        The background is as following:
        For a dataset with face connection,
        when one is finding the neighboring cells for vector interpolation,
        the fact that faces can be rotated against each other can create
        local inconsistency in vector definition near face connections.
        This method corrects that.

        Parameters
        ----------
        faces: iterable
            1D iterable of faces, the first one is assumed to be the reference.
        """
        if self.typ == "LLC":
            return _llc_get_uv_mask_from_face(faces)
        elif self.typ in ["x_periodic", "box"]:
            raise ValueError(
                "It makes no sense to tinker with face_connection when there is only one face"
            )
        else:
            raise NotImplementedError

    def four_matrix_for_uv(self, fface):
        """Expand get_uv_mask_from_face to 2D array of faces."""
        # apply get_uv_mask for the n*m matrix
        UfromUvel, UfromVvel, VfromUvel, VfromVvel = (
            np.zeros(fface.shape) for i in range(4)
        )
        for i in range(fface.shape[0]):
            (
                UfromUvel[i],
                UfromVvel[i],
                VfromUvel[i],
                VfromVvel[i],
            ) = self.get_uv_mask_from_face(fface[i])
        return UfromUvel, UfromVvel, VfromUvel, VfromVvel



import dask.array
import numpy as np


def slice_data_and_shift_indexes(da, indexes_tuple):
    """Slice data using min/max indexes, and shift indexes."""
    slicers = ()
    for indexes, size in zip(indexes_tuple, da.shape):
        if (indexes < 0).any():
            slicers += (slice(None),)
        else:
            start = indexes[indexes >= 0].min()
            stop = stop if (stop := indexes.max() + 1) < size else None
            slicers += (slice(start, stop),)
    indexes_tuple = tuple(
        indexes.ravel() - slicer.start if slicer.start else indexes.ravel()
        for indexes, slicer in zip(indexes_tuple, slicers)
    )
    return da.data[slicers], indexes_tuple


def smart_read(da, indexes_tuple, dask_more_efficient=10, chunks="auto"):
    """Read from a xarray.DataArray given a tuple indexes.

    Try to do it fast and smartly.
    There is a lot of improvement to be made here,
    but this is how it is currently done.

    The data we read is going to be unstructured but they tend to be
    rather localized. For example, the lagrangian particles read data
    from the same time step. Currently, using dask/xarray's unstructured
    read does not really take advantage of the locality.
    This function figures out which chunks stores the data, convert them
    into numpy arrays, and then read the data from the converted ones.

    Parameters
    ----------
    da: xarray.DataArray
        DataArray to read from
    indexes_tuple: tuple of numpy.ndarray
        The indexes of points of interest, each element does not need to be 1D
    dask_more_efficient: int, default 100
        When the number of chunks is larger than this, and the data points are few,
        it may make sense to directly use dask's vectorized read.
    chunks: int, str, default: "auto"
        Chunks for indexes

    Returns
    -------
    + values: numpy.ndarray
        The values of the points of interest. Has the same shape as the elements in indexes_tuple.
    """
    num_dims = len(indexes_tuple)
    if num_dims != da.ndim:
        raise ValueError(
            "indexes_tuple does not match the number of dimensions: "
            f"{len(indexes_tuple)} vs {da.ndim}"
        )

    shape = indexes_tuple[0].shape
    size = indexes_tuple[0].size
    indexes_tuple = tuple(index.ravel() for index in indexes_tuple)
    if not size:
        # This is to make the special case of reading nothing
        # looks normal in other parts of the code.
        return np.empty(shape)

    # data = da.data
    data, indexes_tuple = slice_data_and_shift_indexes(da, indexes_tuple)
    if isinstance(data, np.ndarray):
        return data[indexes_tuple].reshape(shape)

    if dask.array.empty(size, chunks=chunks).numblocks[0] > 1:
        indexes_tuple = tuple(
            dask.array.from_array(indexes, chunks=chunks) for indexes in indexes_tuple
        )

    block_dict = {}
    for block_ids in np.ndindex(*data.numblocks):
        if len(block_dict) >= dask_more_efficient:
            return (
                data.vindex[tuple(map(dask.array.compute, indexes_tuple))]
                .compute()
                .reshape(shape)
            )

        shifted_indexes = []
        mask = None
        for block_id, indexes, chunks in zip(block_ids, indexes_tuple, data.chunks):
            shifted = indexes - sum(chunks[:block_id])
            block_mask = (shifted >= 0) & (shifted < chunks[block_id])
            if not (mask := block_mask if mask is None else mask & block_mask).any():
                break  # empty block
            shifted_indexes.append(shifted)
        else:
            block_dict[block_ids] = (
                np.nonzero(mask),
                tuple(indexes[mask] for indexes in shifted_indexes),
            )
            if sum([len(v[0]) for v in block_dict.values()]) == size:
                break  # all blocks found

    values = np.empty(size)
    for block_ids, (values_indexes, block_indexes) in block_dict.items():
        block_values = data.blocks[block_ids].compute()
        values[values_indexes] = block_values[block_indexes]
    return values.reshape(shape)



import numpy as np
import pytest

import seaduck as sd
import seaduck.kernel_weight as kw


@pytest.fixture
def aknw():
    return kw.KnW(
        inheritance=[[0, 1, 2, 3, 4, 5, 6, 7, 8]],
        hkernel="interp",
        vkernel="linear",
        tkernel="dt",
    )


def test_same_size(aknw):
    assert not aknw.same_size(sd.lagrangian.uknw)


def test_equal(aknw):
    assert not aknw == 1


@pytest.mark.parametrize(
    "masked,ans",
    [
        (
            np.array(
                [
                    [1, 1, 1, 1, 1, 1, 1, 1, 1],
                    [1, 1, 1, 1, 0, 1, 1, 1, 1],
                    [1, 1, 0, 1, 0, 1, 1, 1, 1],
                    [1, 0, 1, 1, 0, 1, 1, 1, 1],
                    [0, 1, 1, 1, 0, 1, 1, 1, 1],
                ]
            ),
            [[0], [1], [2], [3]],
        ),
        (np.array([[0, 1, 1, 1, 0, 1, 1, 1, 1]]), [[] for i in range(4)]),
    ],
)
def test_cascade(masked, ans):
    pk = kw.find_which_points_for_each_kernel(masked)
    assert ans == pk


@pytest.mark.parametrize(
    "rx,ry", [(np.array([0]), np.array([0])), (np.array([0.5]), np.array([0.08]))]
)
@pytest.mark.parametrize(
    "pk,expected",
    [
        ([[], [], [], []], np.nan),
        ([[0], [], [], []], 1),
        ([[], [], [0], []], 1),
    ],
)
def test_cascade_weight(rx, ry, pk, expected):
    actual = np.sum(kw.get_weight_cascade(rx, ry, pk))
    np.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize(
    "rx,ry",
    [
        (np.array([0]), np.array([0])),
        (np.array([0.5]), np.array([0.08])),
    ],
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.array(
            [[0, 0], [0, 1], [0, 2], [0, -1], [0, -2], [-1, 0], [-2, 0], [1, 0], [2, 0]]
        ),
        np.array([[0, 0], [0, 1], [0, -1], [-1, 0], [1, 0]]),
        np.array([[0, 0], [0, -1], [-1, 0], [1, 0]]),
        np.array([[0, 0], [-1, 0], [1, 0]]),
        np.array([[0, 0], [0, 1], [0, -1]]),
        np.array([[0, 0]]),
        np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
    ],
)
def test_interp_func(kernel, rx, ry):
    func = kw.get_func(kernel)
    weight = func(rx, ry)
    assert weight.shape == (1, len(kernel))
    np.testing.assert_allclose(np.sum(weight), 1)


@pytest.mark.parametrize(
    "rx,ry",
    [
        (np.array([0]), np.array([0])),
        (np.array([0.5]), np.array([0.08])),
    ],
)
@pytest.mark.parametrize("kernel_type", ["dx", "dy"])
@pytest.mark.parametrize("horder", [0, 1])
def test_create_different_square(kernel_type, horder, rx, ry):
    k = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
    func = kw.get_func(k, hkernel=kernel_type, h_order=horder)
    weight = func(rx, ry)
    assert weight.shape == (1, len(k))
    if horder == 0:
        np.testing.assert_allclose(np.sum(weight), 1)
    else:
        np.testing.assert_allclose(np.sum(weight), 0, atol=1e-14)


@pytest.mark.parametrize("hkernel", ["dx", "dy"])
def test_auto_inheritance(hkernel):
    k = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
    inheritance = kw._auto_inheritance(k, hkernel=hkernel)
    assert len(inheritance) > 1
    assert isinstance(inheritance[0], list)


@pytest.mark.parametrize(
    "rx,ry", [(np.array([0]), np.array([0])), (np.array([0.5]), np.array([0.08]))]
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.array(
            [[0, 0], [0, 1], [0, 2], [0, -1], [0, -2], [-1, 0], [-2, 0], [1, 0], [2, 0]]
        ),
        np.array([[0, 0], [0, 1], [0, -1], [-1, 0], [1, 0]]),
        np.array([[0, 0], [0, -1], [-1, 0], [1, 0]]),
        np.array([[0, 0], [-1, 0], [1, 0]]),
    ],
)
@pytest.mark.parametrize("order", [1, 2])
def test_dx(kernel, rx, ry, order):
    func = kw.get_func(kernel, hkernel="dx", h_order=order)
    weight = func(rx, ry)
    assert weight.shape == (1, len(kernel))
    assert np.allclose(0, np.sum(weight), atol=1e-14)


@pytest.mark.parametrize(
    "rx,ry", [(np.array([0]), np.array([0])), (np.array([0.5]), np.array([0.08]))]
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.array(
            [[0, 0], [0, 1], [0, 2], [0, -1], [0, -2], [-1, 0], [-2, 0], [1, 0], [2, 0]]
        ),
        np.array([[0, 0], [0, 1], [0, -1], [-1, 0], [1, 0]]),
        np.array([[0, 0], [0, 1], [0, -1]]),
    ],
)
@pytest.mark.parametrize("order", [1, 2])
def test_dy(kernel, rx, ry, order):
    func = kw.get_func(kernel, hkernel="dy", h_order=order)
    weight = func(rx, ry)
    assert weight.shape == (1, len(kernel))
    assert np.allclose(0, np.sum(weight), atol=1e-14)


@pytest.mark.parametrize("kernel_type", ["dx", "dy"])
@pytest.mark.parametrize(
    "kernel,order",
    [
        (
            np.array(
                [
                    [0, 0],
                    [0, 1],
                    [0, 2],
                    [0, -1],
                    [0, -2],
                    [-1, 0],
                    [-2, 0],
                    [1, 0],
                    [2, 0],
                ]
            ),
            5,
        ),
        (
            np.array([[0, 0], [0, 1], [0, -1], [-1, 0], [1, 0]]),
            3,
        ),
        (np.array([[0, 0]]), 1),
        (np.array([[0, 0], [0, 1], [1, 0], [1, 1]]), 2),
    ],
)
def test_order_too_high_error(kernel, order, kernel_type):
    with pytest.raises(Exception):
        kw.kernel_weight(kernel, kernel_type=kernel_type, order=order)


def test_plot_kernel():
    pytest.importorskip("matplotlib.pyplot")
    kw.show_kernels()
    # no way to assert here


def test_dt_and_bottom_scheme(aknw):
    weight = aknw.get_weight(np.array([0.618]), np.array([0.0618]))
    assert weight.shape[0] == 1
    assert np.allclose(0, np.sum(weight))


@pytest.mark.parametrize("which", ["dx", "dy"])
def test_maxorder_dxdy_square(which):
    kkk = np.array(
        [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2], [0, 2], [2, 2], [2, 0], [2, 1]]
    )
    xorder = 0
    yorder = 0
    if which == "dx":
        xorder = 2
    else:
        yorder = 2
    func = kw.kernel_weight_s(kkk, xorder, yorder)
    weight = func(np.array([0.0]), np.array([0.0]))
    assert weight.shape == (1, len(kkk))
    assert np.allclose(0, np.sum(weight), atol=1e-14)



import numpy as np
import pytest

from seaduck.topology import Topology, _llc_get_uv_mask_from_face, _llc_mutual_direction

try:
    import numba

    print(numba)
    # just to avoid the line got deleted by pre-commit
    HAS_NUMBA = True
except ImportError:
    HAS_NUMBA = False


@pytest.mark.parametrize("face", [1, 2, 4, 5, 6, 7, 8, 10, 11])
@pytest.mark.parametrize("edge", [0, 1, 2, 3])
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_get_the_neighbor_face(tp, face, edge):
    nf, ne = tp.get_the_other_edge(face, edge)
    assert nf in range(13)
    assert ne in range(4)


@pytest.mark.parametrize(
    "typ,error",
    [
        ("box", Exception),
        ("x_periodic", Exception),
        ("cubed_sphere", NotImplementedError),
        ("other", NotImplementedError),
    ],
)
@pytest.mark.parametrize(
    "func,args",
    [
        ("get_the_other_edge", (0, 0)),
        ("mutual_direction", (0, 1)),
    ],
)
@pytest.mark.parametrize("ds", ["ecco"], indirect=True)
def test_not_applicable(ds, typ, func, args, error):
    tpp = Topology(ds, typ)
    with pytest.raises(error):
        getattr(tpp, func)(*args)


@pytest.mark.parametrize("face,edge", [(0, 1), (3, 1), (9, 3), (12, 3)])
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_antarctica_error(tp, face, edge):
    with pytest.raises(IndexError):
        tp.get_the_other_edge(face, edge)


@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_mutual_face(tp):
    e1, _ = tp.mutual_direction(0, 1)
    assert e1 in [0, 1, 2, 3]


@pytest.mark.parametrize(
    "ind,tend,result",
    [
        ((1, 45, 45), 0, (1, 46, 45)),
        ((1, 45, 45), 1, (1, 44, 45)),
        ((1, 45, 45), 2, (1, 45, 44)),
        ((1, 45, 45), 3, (1, 45, 46)),
        ((5, 89, 89), 3, (7, 0, 0)),
        ((5, 89, 89), 0, (6, 0, 89)),
        ((6, 89, 89), 0, (10, 0, 0)),
        ((6, 0, 0), 2, (2, 89, 89)),
    ],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_llc_ind_tend(tp, ind, tend, result):
    res = tp.ind_tend(ind, tend)
    assert res == result


@pytest.mark.parametrize("tend", [0, 1, 2, 3])
@pytest.mark.parametrize("tp", ["aviso", "rect"], indirect=True)
def test_boxish_ind_tend(tp, tend):
    ind = tp.ind_tend((0, 0), tend)
    assert len(ind) == 2


@pytest.mark.parametrize("tend", [0, 3])
@pytest.mark.parametrize("tp", ["aviso", "rect"], indirect=True)
def test_boundary_case(tp, tend):
    out = tp.ind_tend((tp.ixmax, tp.iymax), tend)
    if tp.typ == "box" or tend == 0:
        assert -1 in out

    out = tp.ind_tend((-1, -1), tend)
    assert -1 in out


mundane = np.array([[[1.0, 1.0]], [[0.0, 0.0]], [[0.0, -0.0]], [[1.0, 1.0]]])


@pytest.mark.parametrize(
    "fface,cis",
    [
        (np.array([[1, 1]]), True),
        (np.array([[1, 2]]), True),
        (np.array([[10, 2]]), False),
        (np.array([[6, 10]]), False),
    ],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_4_matrix(tp, fface, cis):
    ans = np.array(tp.four_matrix_for_uv(fface))
    if cis:
        assert np.allclose(ans, mundane)
    else:
        assert not np.allclose(ans, mundane)


@pytest.mark.parametrize("ds", ["rect"], indirect=True)
def test_unable_to_cereate(ds):
    temp = ds.drop_vars("XC")
    with pytest.raises(KeyError):
        Topology(temp)


@pytest.mark.parametrize("ds", ["aviso"], indirect=True)
def test_create_without_time(ds):
    temp = ds.drop_vars("time")
    tp = Topology(temp)
    assert tp.itmax == 0


@pytest.mark.parametrize(
    "func,args,kwargs,error",
    [
        ("ind_tend", ((1, 45, 45), 0), {"cuvwg": "other"}, ValueError),
        ("ind_moves", ((1, 45, 45), ["left", "left"]), {}, ValueError),
    ],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_other_errors(tp, func, args, kwargs, error):
    with pytest.raises(error):
        getattr(tp, func)(*args, **kwargs)


@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_ind_moves_with1illegal(tp):
    res = tp.ind_moves((1, -1, 89), [0, 0])
    assert res == (-1, -1, -1)


@pytest.mark.parametrize(
    "ind,tend,ans",
    [
        ((1, 45, 45), 0, (1, 46, 45)),
        ((1, 45, 45), 2, (1, 45, 44)),
        ((1, 0, 0), 2, (12, 89, 0)),
        ((1, 0, 0), 1, (0, 89, 0)),
        ((4, 45, 89), 3, (8, 0, 45)),
    ],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_ind_tend_v(tp, ind, tend, ans):
    res = tp.ind_tend(ind, tend, cuvwg="V")
    assert res == ans


@pytest.mark.parametrize(
    "ans,tend,ind",
    [
        ((1, 45, 45), 1, (1, 46, 45)),
        ((1, 45, 44), 2, (1, 45, 45)),
        ((1, 0, 0), 0, (12, 89, 0)),
        ((4, 45, 89), 1, (8, 0, 45)),
    ],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_ind_tend_u(tp, ind, tend, ans):
    res = tp.ind_tend(ind, tend, cuvwg="U")
    assert res == ans


@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_wall_between(tp):
    # it is a bit hard to think about an example
    # that uses this case from higher level.
    uv, R = tp._find_wall_between((11, 0, 45), (8, 89, 45))
    assert uv == "V"
    assert R == (11, 0, 45)
    # It does not make sense to parametrize now.
    uv, R = tp._find_wall_between((8, 45, 0), (7, 45, 89))
    assert uv == "U"
    assert R == (8, 45, 0)


@pytest.mark.parametrize(
    "face,nface,rot",
    [
        (4, 9, True),
        (3, 8, True),
        (9, 4, True),
        (8, 3, True),
        (0, 4, False),
        (1, 3, False),
    ],
)
def test_transitive_mututal_direction(face, nface, rot):
    e1, e2 = _llc_mutual_direction(face, nface, transitive=True)
    isrot = (e1 // 2 + e2 // 2) % 2
    assert isrot == rot


@pytest.mark.skipif(
    HAS_NUMBA, reason="Weird numba behavior that needs to be revisited."
)
@pytest.mark.parametrize("transitive,face, nface", [(True, 0, 7), (False, 4, 9)])
def test_mutual_face_error(transitive, face, nface):
    with pytest.raises(ValueError):
        _llc_mutual_direction(face, nface, transitive=transitive)


def test_uv_mask():
    faces = np.array([1, 1, 1, 1, 4])
    uu, uv, vu, vv = _llc_get_uv_mask_from_face(faces)
    assert (uu == 1).all()


@pytest.mark.parametrize("tp", ["rect"], indirect=True)
def test_uv_mask_error(tp):
    faces = np.array([1, 1, 1, 1, 4])
    with pytest.raises(Exception):
        tp.get_uv_mask_from_face(faces)


@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_wall_between_itself(tp):
    with pytest.raises(IndexError):
        tp._find_wall_between((1, 14, 14), (1, 14, 14))


@pytest.mark.parametrize(
    "ind, tend, exp",
    [((10, 89, 45), 0, ("U", (2, 44, 0))), ((7, 0, 45), 1, ("U", (5, 44, 89)))],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_ind_tend_v_detailed(tp, ind, tend, exp):
    ans = tp._ind_tend_V(ind, tend)
    assert ans == exp


@pytest.mark.parametrize(
    "ind, tend, exp",
    [((2, 44, 0), 2, ("V", (10, 89, 45))), ((5, 44, 89), 3, ("V", (7, 0, 45)))],
)
@pytest.mark.parametrize("tp", ["ecco"], indirect=True)
def test_ind_tend_u_detailed(tp, ind, tend, exp):
    ans = tp._ind_tend_U(ind, tend)
    assert ans == exp



import copy
import logging

import numpy as np

from seaduck.get_masks import get_masked
from seaduck.kernel_weight import KnW, _find_pk_4d, _translate_to_tendency
from seaduck.ocedata import HRel, OceData, RelCoord, TRel, VlLinRel, VlRel, VRel
from seaduck.smart_read import smart_read
from seaduck.utils import (
    _general_len,
    _read_h,
    find_px_py,
    get_key_by_value,
    local_to_latlon,
    to_180,
    weight_f_node,
)


def _ind_broadcast(x, ind):
    """Perform a "cartesian product" on two fattened dimensions.

    Parameters
    ----------
    x: numpy.ndarray
        A new dimension
    ind: tuple
        Existing indexes
    """
    n = x.shape[0]
    if len(x.shape) == 1:
        x = x.reshape((n, 1))
    xsp = x.shape
    ysp = ind[0].shape
    final_shape = [n] + list(ysp[1:]) + list(xsp[1:])

    to_return = [np.zeros(final_shape, int) for i in range(len(ind) + 1)]

    dims = len(final_shape)
    ydim = len(ysp) - 1
    trsp = list(range(1, 1 + ydim)) + [0] + list(range(1 + ydim, dims))
    inv = np.argsort(trsp)
    to_return[0] = to_return[0].transpose(trsp)
    to_return[0][:] = x
    to_return[0] = to_return[0].transpose(inv)

    for i in range(1, len(ind) + 1):
        to_return[i] = to_return[i].T
        to_return[i][:] = ind[i - 1].T
        to_return[i] = to_return[i].T
    return to_return


def _partial_flatten(ind):
    """Convert a high dimensional index set to a 2D one."""
    if isinstance(ind, tuple):
        shape = ind[0].shape

        # num_neighbor = 1
        # for i in range(1,len(shape)):
        #     num_neighbor*=shape[i]
        to_return = []
        for i in range(len(ind)):
            to_return.append(ind[i].reshape(shape[0], -1))
        return tuple(to_return)
    elif isinstance(ind, np.ndarray):
        shape = ind.shape
        num_neighbor = 1
        for i in range(1, len(shape)):
            num_neighbor *= shape[i]
        return ind.reshape(shape[0], num_neighbor)
    else:
        raise NotImplementedError("ind type not supported")


def _in_required(name, required):
    """See if a name is in required."""
    if required == "all":
        return True
    else:
        return name in required


def _ind_for_mask(ind, dims):
    """Find the index for masking.

    If dims does not include a vertical dimension, assume to be 0.
    If dims has a temporal dimension, take it away.
    Return the index for masking.
    """
    ind_for_mask = [ind[i] for i in range(len(ind)) if dims[i] not in ["time"]]
    if "Z" not in dims and "Zl" not in dims:
        ind_for_mask.insert(0, np.zeros_like(ind[0]))
    return tuple(ind_for_mask)


def _subtract_prefetch_prefix(ind, prefetch_prefix):
    """Subtract the index prefix from the actual index.

    This is used when one is reading from a prefetched subset of the data.
    """
    temp_ind = []
    for i in range(len(prefetch_prefix)):
        temp_ind.append(ind[i] - prefetch_prefix[i])
    return tuple(temp_ind)


class Position:
    """The Position object that performs the interpolation.

    Create a empty one by default.
    To actually do interpolation, use from_latlon method to tell the ducks where they are.
    """

    def __new__(cls, *arg, **kwarg):
        new_position = object.__new__(cls)
        new_position.rel = RelCoord()
        return new_position

    def __getattr__(self, attr):
        if attr == "rel":
            object.__getattribute__(self, attr)
        elif attr in self.rel.keys():
            return getattr(self.rel, attr)
        else:
            object.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        if attr == "rel":
            object.__setattr__(self, attr, value)
        elif attr in self.rel.keys():
            setattr(self.rel, attr, value)
        else:
            object.__setattr__(self, attr, value)

    def from_latlon(self, x=None, y=None, z=None, t=None, data=None):
        """Fill in the coord info using lat-lon-dep-time dims.

        Use the methods from the ocedata to transform
        from lat-lon-dep-time coords to rel-coords
        store the output in the Position object.

        Parameters
        ----------
        x,y,z,t: numpy.ndarray, float or None, default None
            1D array of the lat-lon-dep-time coords
        data: OceData object
            The field where the Positions are defined on.
        """
        if data is None:
            try:
                self.ocedata
            except AttributeError as exc:
                raise ValueError("data not provided") from exc
        else:
            if isinstance(data, OceData):
                self.ocedata = data
            else:
                raise ValueError("Input data must be OceData")
        self.tp = self.ocedata.tp
        length = [_general_len(i) for i in [x, y, z, t]]
        self.N = max(length)
        if any(i != self.N for i in length if i > 1):
            raise ValueError("Shapes of input coordinates are not compatible")

        if isinstance(x, (int, float, np.floating)):
            x = np.ones(self.N, float) * x
        if isinstance(y, (int, float, np.floating)):
            y = np.ones(self.N, float) * y
        if isinstance(z, (int, float, np.floating)):
            z = np.ones(self.N, float) * z
        if isinstance(t, (int, float, np.floating)):
            t = np.ones(self.N, float) * t

        for thing in [x, y, z, t]:
            if thing is None:
                continue
            if len(thing.shape) > 1:
                raise ValueError("Input need to be 1D numpy arrays")
        if (x is not None) and (y is not None):
            self.lon = copy.deepcopy(x)
            self.lat = copy.deepcopy(y)
            self.rel.update(self.ocedata._find_rel_h(self.lon, self.lat))
        else:
            self.rel.update(HRel._make([None for i in range(11)]))
            self.lon = None
            self.lat = None
        if z is not None:
            self.dep = copy.deepcopy(z)
            if self.ocedata.readiness["Z"]:
                self.rel.update(self.ocedata._find_rel_v(z))
            else:
                self.rel.update(VRel._make(None for i in range(4)))
            if self.ocedata.readiness["Zl"]:
                self.rel.update(self.ocedata._find_rel_vl(z))
                self.rel.update(self.ocedata._find_rel_vl_lin(z))
            else:
                self.rel.update(VlRel._make(None for i in range(4)))
                self.rel.update(VlLinRel._make(None for i in range(4)))
        else:
            self.rel.update(VRel._make(None for i in range(4)))
            self.rel.update(VlRel._make(None for i in range(4)))
            self.rel.update(VlLinRel._make(None for i in range(4)))
            self.dep = None
        if t is not None:
            self.t = copy.deepcopy(t)
            if self.ocedata.readiness["time"]:
                self.rel.update(self.ocedata._find_rel_t(t))
            else:
                self.rel.update(TRel._make(None for i in range(4)))
        else:
            self.rel.update(TRel._make(None for i in range(4)))
            self.t = None
            # (self.it, self.rt, self.dt, self.bt, self.t) = (None for i in range(5))
        return self

    def from_bool_array(
        self, t=None, data=None, bool_array=None, num=None, random_seed=None
    ):
        """Update/Generate new object with random points in given grid boxes.

        Use the methods from the ocedata to transform
        from lat-lon-dep-time coords to rel-coords
        store the output in the Position object.

        Parameters
        ----------
        t: numpy.ndarray, float or None, default None
            1D array of the time coords
        data: OceData object
            The field where the Positions are defined on.
        bool_array: numpy.ndarray, or xr.DataArray
            Points are generated where it is True.
        num: int
            Total number of particles to seed (approximately).
        random_seed: int optional
            The random seed used for reproducible results.
        """
        if random_seed is not None:
            np.random.seed(random_seed)

        if isinstance(data, OceData):
            if data.readiness["h"] != "oceanparcel":
                raise NotImplementedError(
                    "This method only support datasets that has XG, YG in it."
                )
            if not (data.readiness["Z"] and data.readiness["Zl"]):
                raise NotImplementedError(
                    "This method only support datasets that has all Z coords."
                )
            self.ocedata = data
        else:
            raise ValueError("Input data must be OceData")
        try:
            self.ocedata["Vol"]
        except KeyError:
            self.ocedata._add_missing_vol()
        self.tp = self.ocedata.tp

        inds = np.where(bool_array)
        try:
            data["Vol"]
        except KeyError:
            data._add_missing_vol()
        np_vol = np.array(data["Vol"])
        vols = np.array(np_vol[inds])
        num_each = np.round(vols * num / np.sum(vols)).astype(int)
        num = np.sum(num_each)
        self.N = num
        inds = tuple(np.repeat(np.array(i), num_each) for i in inds)
        if len(inds) == 3:
            iz, iy, ix = inds
            ind = (iy, ix)
            face = None
        elif len(inds) == 4:
            iz, face, iy, ix = inds
            ind = (face, iy, ix)
        else:
            raise ValueError("bool_array must be 3 or 4 dimensional")
        cs, sn, dx, dy, bx, by = _read_h(
            data.XC, data.YC, data.dX, data.dY, data.CS, data.SN, ind
        )
        rx = np.random.random(num) - 0.5
        ry = np.random.random(num) - 0.5
        rzl_lin = np.random.random(num)
        self.rel.update(HRel(face, iy, ix, rx, ry, cs, sn, dx, dy, bx, by))

        izl_lin = iz + 1
        bzl_lin = data.Zl[izl_lin]
        dzl_lin = data.dZl[izl_lin]
        self.rel.update(VlLinRel(izl_lin, rzl_lin, dzl_lin, bzl_lin))

        # set lon temporaily
        self.lon = bx
        px, py = self.get_px_py()
        w = self.get_f_node_weight()
        self.lon = np.einsum("nj,nj->n", w, px.T)
        self.lat = np.einsum("nj,nj->n", w, py.T)

        self.dep = self.bzl_lin + self.dzl_lin * self.rzl_lin
        self.rel.update(self.ocedata._find_rel_v(self.dep))
        self.rel.update(self.ocedata._find_rel_vl(self.dep))

        if isinstance(t, (int, float, np.floating)):
            t = np.ones(self.N, float) * t
        elif isinstance(t, np.ndarray):
            if len(t) != self.N:
                raise ValueError("Mismatch between input t and final particle number")
        if t is not None:
            self.t = copy.deepcopy(t)
            if self.ocedata.readiness["time"]:
                self.rel.update(self.ocedata._find_rel_t(t))
            else:
                self.rel.update(TRel._make(None for i in range(4)))
        else:
            self.rel.update(TRel._make(None for i in range(4)))
            self.t = None
        return self

    def subset(self, which):
        """Create a subset of the Position object.

        Parameters
        ----------
        which: numpy.ndarray, optional
            Define which points survive the subset operation.
            It be an array of either boolean or int.
            The selection is similar to that of selecting from a 1D numpy array.

        Returns
        -------
        the_subset: Position object
            The selected Positions.
        """
        p = object.__new__(type(self))
        vardict = vars(self)
        keys = vardict.keys()
        for key in keys:
            item = vardict[key]
            if isinstance(item, np.ndarray):
                if len(item.shape) == 1:
                    setattr(p, key, item[which])
                    p.N = len(getattr(p, key))
                elif key in ["px", "py"]:
                    setattr(p, key, item[:, which])
                else:
                    setattr(p, key, item)
            elif isinstance(item, RelCoord):
                setattr(p, key, item.subset(which))
            else:
                setattr(p, key, item)
        return p

    def update_from_subset(self, sub, which):
        """Update from the original one from a subset of the Position object.

        Parameters
        ----------
        sub: Position object
            The Position object to be updated from.
        which: numpy.ndarray, optional
            Define which points correpond to the subset
            It be an array of either boolean or int.
            The selection is similar to that of selecting from a 1D numpy array.
        """
        vardict = vars(sub)
        keys = vardict.keys()
        for key in keys:
            item = vardict[key]
            if not hasattr(self, key):
                logging.warning(
                    f"A new attribute {key} defined" "after updating from subset"
                )
                setattr(self, key, item)
            if getattr(self, key) is None:
                continue
            if isinstance(item, np.ndarray):
                if len(item.shape) == 1:
                    getattr(self, key)[which] = item
                elif key in ["px", "py"]:
                    getattr(self, key)[:, which] = item
            elif isinstance(item, RelCoord):
                self.rel.update_from_subset(item, which)
            elif isinstance(item, list):
                setattr(self, key, item)

    def _fatten_h(self, knw, ind_moves_kwarg={}):
        """Fatten horizontal indices.

        Fatten means to find the neighboring points of the points of interest based on the kernel.
        faces,iys,ixs are 1d arrays of size n.
        We are applying a kernel of size m.
        This is going to return a n * m array of indexes.
        A very slim vector is now a matrix, and hence the name.
        each row represen all the node needed for interpolation of a single point.
        "h" represent we are only doing it on the horizontal plane.

        Parameters
        ----------
        knw: KnW object
            The kernel used to find neighboring points.
        ind_moves_kward: dict, optional
            Key word argument to put into ind_moves method of the Topology object.
            Read Topology.ind_moves for more detail.
        """
        #         self.ind_h_dict
        kernel = knw.kernel.astype(int)
        kernel_tends = [_translate_to_tendency(k) for k in kernel]
        m = len(kernel_tends)
        n = len(self.iy)
        tp = self.ocedata.tp

        # the arrays we are going to return
        if self.face is not None:
            n_faces = np.zeros((n, m), int)
            n_faces.T[:] = self.face
        n_iys = np.zeros((n, m), int)
        n_ixs = np.zeros((n, m), int)

        # first try to fatten it naively(fast and vectorized)
        for i, node in enumerate(kernel):
            x_disp, y_disp = node
            n_iys[:, i] = self.iy + y_disp
            n_ixs[:, i] = self.ix + x_disp
        cuvwg = ind_moves_kwarg.get("cuvwg", "C")
        if self.face is not None:
            illegal = tp.check_illegal((n_faces, n_iys, n_ixs), cuvwg=cuvwg)
        else:
            illegal = tp.check_illegal((n_iys, n_ixs), cuvwg=cuvwg)

        redo = np.array(np.where(illegal)).T
        for loc in redo:
            j, i = loc
            if self.face is not None:
                ind = (self.face[j], self.iy[j], self.ix[j])
            else:
                ind = (self.iy[j], self.ix[j])
            # everyone start from the [0,0] node
            moves = kernel_tends[i]
            # moves is a list of operations to get to a single point
            # [2,2] means move to the left and then move to the left again.
            n_ind = tp.ind_moves(ind, moves, **ind_moves_kwarg)
            if self.face is not None:
                n_faces[j, i], n_iys[j, i], n_ixs[j, i] = n_ind
            else:
                n_iys[j, i], n_ixs[j, i] = n_ind
        if self.face is not None:
            return n_faces, n_iys, n_ixs
        else:
            return None, n_iys, n_ixs

    def _fatten_v(self, knw):
        """Fatten in vertical center coord.

        Find the neighboring center grid points in the vertical direction.

        Parameters
        ----------
        knw: KnW object
            The kernel used to find neighboring points.
        """
        if self.iz is None:
            return None
        if knw.vkernel == "nearest":
            return copy.deepcopy(self.iz)
        elif knw.vkernel in ["dz", "linear"]:
            try:
                self.iz_lin
            except AttributeError:
                self.rel.update(self.ocedata._find_rel_v_lin(self.dep))
            return np.vstack([self.iz_lin, self.iz_lin - 1]).T
        else:
            raise ValueError("vkernel not supported")

    def _fatten_vl(self, knw):
        """Fatten in vertical staggered coord.

        Finding the neighboring staggered grid points in the vertical direction.

        Parameters
        ----------
        knw: KnW object
            The kernel used to find neighboring points.
        """
        if self.izl is None:
            return None
        if knw.vkernel == "nearest":
            return copy.deepcopy(self.izl)
        elif knw.vkernel in ["dz", "linear"]:
            try:
                self.izl_lin
            except AttributeError:
                self.rel.update(self.ocedata._find_rel_vl_lin(self.dep))
            return np.vstack([self.izl_lin, self.izl_lin - 1]).T
        else:
            raise ValueError("vkernel not supported")

    def _fatten_t(self, knw):
        """Fatten in the temporal coord.

        Finding the neighboring center grid points in the temporal dimension.

        Parameters
        ----------
        knw: KnW object
            The kernel used to find neighboring points.
        """
        if self.it is None:
            return None
        if knw.tkernel == "nearest":
            return copy.deepcopy(self.it)
        elif knw.tkernel in ["dt", "linear"]:
            try:
                self.it_lin
            except AttributeError:
                self.rel.update(self.ocedata._find_rel_t_lin(self.t))
            return np.vstack([self.it_lin, self.it_lin + 1]).T
        else:
            raise ValueError("tkernel not supported")

    def fatten(self, knw, four_d=False, required="all", ind_moves_kwarg={}):
        """Fatten in all the required dimensions.

        Finding the neighboring center grid points in all 4 dimensions.

        Parameters
        ----------
        knw: KnW object
            The kernel used to find neighboring points.
        four_d: Boolean, default False
            When we are doing nearest neighbor interpolation on some of the dimensions,
            with four_d = True, this will create dimensions with length 1, and will squeeze
            the dimension if four_d = False
        required: str, iterable of str, default "all"
            Which dims is needed in the process
        ind_moves_kward: dict, optional
            Key word argument to put into ind_moves method of the Topology object.
            Read Topology.ind_moves for more detail.
        """
        if required != "all" and isinstance(required, str):
            required = tuple([required])
        if required == "all" or isinstance(required, tuple):
            pass
        else:
            required = tuple(required)

        # TODO: register the kernel shape
        if (
            _in_required("X", required)
            or _in_required("Y", required)
            or _in_required("face", required)
        ):
            ffc, fiy, fix = self._fatten_h(knw, ind_moves_kwarg=ind_moves_kwarg)
            if ffc is not None:
                to_return = (ffc, fiy, fix)
                keys = ["face", "Y", "X"]
            else:
                to_return = (fiy, fix)
                keys = ["Y", "X"]
        else:
            to_return = tuple([np.zeros(self.N)])
            keys = ["place_holder"]

        if _in_required("Z", required):
            fiz = self._fatten_v(knw)
            if fiz is not None:
                to_return = _ind_broadcast(fiz, to_return)
                keys.insert(0, "Z")
        elif _in_required("Zl", required):
            fizl = self._fatten_vl(knw)
            if fizl is not None:
                to_return = _ind_broadcast(fizl, to_return)
                keys.insert(0, "Zl")
        elif four_d:
            to_return = [
                np.expand_dims(to_return[i], axis=-1) for i in range(len(to_return))
            ]

        if _in_required("time", required):
            fit = self._fatten_t(knw)
            if fit is not None:
                to_return = _ind_broadcast(fit, to_return)
                keys.insert(0, "time")
        elif four_d:
            to_return = [
                np.expand_dims(to_return[i], axis=-1) for i in range(len(to_return))
            ]
        to_return = dict(zip(keys, to_return))
        if required == "all":
            required = [i for i in keys if i != "place_holder"]
        return tuple(to_return[i] for i in required)

    def get_px_py(self):
        """Get the nearest 4 corner points of the given point.

        Used for oceanparcel style horizontal interpolation.

        Returns
        -------
        px: numpy.ndarray
            the longitude of the Position's surrounding 4 corner points.
        py: numpy.ndarray
            the latitude of those points mentioned above.
        """
        if self.face is not None:
            ind = (
                self.face,
                self.iy,
                self.ix,
            )
        else:
            ind = (self.iy, self.ix)
        px, py = find_px_py(
            self.ocedata.XG,
            self.ocedata.YG,
            self.ocedata.tp,
            ind,
        )
        px = self.lon + to_180(px - self.lon)
        return px, py

    def get_f_node_weight(self):
        """Find weight for the corner points interpolation."""
        return weight_f_node(self.rx, self.ry)

    def _register_interpolation_input(
        self, var_name, knw, prefetched=None, prefetch_prefix=None
    ):
        """Register the input of interpolation function.

        Part of the interpolation function.
        Register the input of the interpolation function.
        For the input format, go to interpolation for more details.

        Returns
        -------
        output_format: dict
            Record information about the original var_name input.
            Provide the formatting information for output,
            such that it matches the input in a direct fashion.
        main_keys: list
            A list that defines each interpolation to be performed.
            The combination of variable name and kernel uniquely define such an operation.
        prefetch_dict: dict
            Lookup the prefetched the data and its index prefix using main_key
        main_dict: dict
            Lookup the raw information using main_key
        hash_index: dict
            Lookup the token that uniquely define its indexing operation using main_key.
            Different main_key could share the same token.
        hash_mask: dict
            Lookup the token that uniquely define its masking operation using main_key.
            Different main_key could share the same token.
        hash_read: dict
            Lookup the token that uniquely define its reading of the data using main_key.
            Different main_key could share the same token.
        hash_weight: dict
            Lookup the token that uniquely define its computation of the weight using main_key.
            Different main_key could share the same token.
        """
        # prefetch_dict = {var:(prefetched,prefetch_prefix)}
        # main_dict = {var:(var,kernel)}
        # hash_index = {var:hash(cuvwg,kernel_size)}
        # hash_read  = {var:hash(var,kernel_size)}
        # hash_weight= {var:hash(kernel,cuvwg)}
        output_format = {}
        if isinstance(var_name, (str, tuple)):
            var_name = [var_name]
            output_format["single"] = True
        elif isinstance(var_name, list):
            output_format["single"] = False
        else:
            raise ValueError(
                "var_name needs to be string, tuple, or a list of the above."
            )
        num_var = len(var_name)

        if isinstance(knw, KnW):
            knw = [knw for i in range(num_var)]
        if isinstance(knw, tuple):
            if len(knw) != 2:
                raise ValueError(
                    "When knw is a tuple, we assume it to be kernels for a horizontal vector,"
                    "thus, it has to have 2 elements"
                )
            knw = [knw for i in range(num_var)]
        elif isinstance(knw, list):
            if len(knw) != num_var:
                raise ValueError("Mismatch between the number of kernels and variables")
        elif isinstance(knw, dict):
            temp = []
            for var in var_name:
                a_knw = knw.get(var)
                if a_knw is None or not isinstance(a_knw, (tuple, KnW)):
                    raise ValueError(
                        f"Variable {var} does not have a proper corresponding kernel"
                    )
                else:
                    temp.append(a_knw)
            knw = temp
        else:
            raise ValueError(
                "knw needs to be a KnW object, or list/dictionaries containing that "
            )

        if isinstance(prefetched, np.ndarray):
            prefetched = [prefetched for i in range(num_var)]
        elif isinstance(prefetched, tuple):
            prefetched = [prefetched for i in range(num_var)]
        elif prefetched is None:
            prefetched = [prefetched for i in range(num_var)]
        elif isinstance(prefetched, list):
            if len(prefetched) != num_var:
                raise ValueError(
                    "Mismatch between the number of prefetched arrays and variables"
                )
        elif isinstance(prefetched, dict):
            prefetched = [prefetched.get(var) for var in var_name]
        else:
            raise ValueError(
                "prefetched needs to be a numpy array/tuple pair of numpy array,"
                " or list/dictionaries containing that"
            )

        if isinstance(prefetch_prefix, tuple):
            prefetch_prefix = [prefetch_prefix for i in range(num_var)]
        elif prefetch_prefix is None:
            prefetch_prefix = [None for i in range(num_var)]
        elif isinstance(prefetch_prefix, list):
            if len(prefetch_prefix) != num_var:
                raise ValueError(
                    "Mismatch between the number of prefetched arrays prefix prefetch_prefix and variables"
                )
        elif isinstance(prefetch_prefix, dict):
            prefetch_prefix = [prefetch_prefix.get(var) for var in var_name]
        else:
            raise ValueError(
                "prefetched prefix prefetch_prefix needs to be a tuple, or list/dictionaries containing that "
            )

        output_format["ori_list"] = copy.deepcopy(list(zip(var_name, knw)))
        new_var_name = []
        new_prefetched = []
        new_knw = []
        new_prefetch_prefix = []
        for i, var in enumerate(var_name):
            if isinstance(var, str):
                new_var_name.append(var)
                new_prefetched.append(prefetched[i])
                new_knw.append(knw[i])
                new_prefetch_prefix.append(prefetch_prefix[i])
            elif isinstance(var, tuple):
                if self.face is None:
                    for j in range(len(var)):
                        new_var_name.append(var[j])
                        if prefetched[i] is not None:
                            new_prefetched.append(prefetched[i][j])
                        else:
                            new_prefetched.append(None)
                        new_knw.append(knw[i][j])
                        new_prefetch_prefix.append(prefetch_prefix[i])
                else:
                    new_var_name.append(var)
                    new_prefetched.append(prefetched[i])
                    new_knw.append(knw[i])
                    new_prefetch_prefix.append(prefetch_prefix[i])
            elif var is None:
                pass
            else:
                raise ValueError(
                    "var_name needs to be string, tuple, or a list of the above."
                )

        prefetched = new_prefetched
        knw = new_knw
        prefetch_prefix = new_prefetch_prefix
        var_name = new_var_name
        output_format["final_var_name"] = list(zip(var_name, knw))

        kernel_size_hash = []
        kernel_hash = []
        mask_ignore = []
        for kkk in knw:
            if isinstance(kkk, KnW):
                kernel_size_hash.append(kkk.size_hash())
                kernel_hash.append(hash(kkk))
                mask_ignore.append(kkk.ignore_mask)
            elif isinstance(kkk, tuple):
                if len(kkk) != 2:
                    raise ValueError(
                        "When knw is a tuple, we assume it to be kernels for a horizontal vector,"
                        "thus, it has to have 2 elements"
                    )
                uknw, vknw = kkk
                kernel_size_hash.append(uknw.size_hash())
                kernel_hash.append(hash((uknw, vknw)))
                mask_ignore.append(uknw.ignore_mask or vknw.ignore_mask)
        dims = []
        for var in var_name:
            if isinstance(var, str):
                dims.append(self.ocedata[var].dims)
            elif isinstance(var, tuple):
                temp = []
                for vvv in var:
                    temp.append(self.ocedata[vvv].dims)
                dims.append(tuple(temp))

        main_keys = list(zip(var_name, kernel_hash))
        prefetch_dict = dict(zip(main_keys, zip(prefetched, prefetch_prefix)))
        main_dict = dict(zip(main_keys, zip(var_name, dims, knw)))
        hash_index = dict(
            zip(main_keys, [hash(i) for i in zip(dims, kernel_size_hash)])
        )
        hash_mask = dict(
            zip(main_keys, [hash(i) for i in zip(dims, mask_ignore, kernel_size_hash)])
        )
        hash_read = dict(
            zip(main_keys, [hash(i) for i in zip(var_name, kernel_size_hash)])
        )
        hash_weight = dict(zip(main_keys, [hash(i) for i in zip(dims, kernel_hash)]))
        return (
            output_format,
            main_keys,
            prefetch_dict,
            main_dict,
            hash_index,
            hash_mask,
            hash_read,
            hash_weight,
        )

    def _fatten_required_index_and_register(self, hash_index, main_dict):
        """Fatten for the interpolation process.

        Perform the fatten process for each unique token. Register the result as a dict.

        Parameters
        ----------
        hash_index: dict
            See _register_interpolation_input
        main_dict: dict
            See _register_interpolation_input

        Returns
        -------
        + index_lookup: dict
            A dictionary to lookup fatten results.
            The keys are the token in hash_index.
            The values are corresponding results.
        """
        hsh = np.unique(list(hash_index.values()))
        index_lookup = {}
        for hs in hsh:
            main_key = get_key_by_value(hash_index, hs)
            var_name, dims, knw = main_dict[main_key]
            if isinstance(var_name, str):
                old_dims = copy.deepcopy(dims)
            elif isinstance(var_name, tuple):
                old_dims = copy.deepcopy(dims[0])
            dims = []
            for i in old_dims:
                if i in ["Xp1", "Yp1"]:
                    dims.append(i[:1])
                else:
                    dims.append(i)
            dims = tuple(dims)
            if "Xp1" in old_dims or "Yp1" in old_dims:
                cuvwg = "G"
            else:
                cuvwg = "C"
            if isinstance(var_name, str):
                ind = self.fatten(
                    knw, required=dims, four_d=True, ind_moves_kwarg={"cuvwg": cuvwg}
                )
                index_lookup[hs] = ind
            elif isinstance(var_name, tuple):
                uknw, vknw = knw
                uind = self.fatten(
                    uknw, required=dims, four_d=True, ind_moves_kwarg={"cuvwg": "U"}
                )
                vind = self.fatten(
                    vknw, required=dims, four_d=True, ind_moves_kwarg={"cuvwg": "V"}
                )
                index_lookup[hs] = (uind, vind)

        return index_lookup

    def _transform_vector_and_register(self, index_lookup, hash_index, main_dict):
        """Transform vectors for interpolation.

        Perform the vector transformation.
        This is to say that sometimes u velocity becomes v velocity for datasets with face topology.
          Register the result as a dict.

        Parameters
        ----------
        index_lookup: dict
            See _fatten_required_index_and_register
        hash_index: dict
            See _register_interpolation_input
        main_dict: dict
            See _register_interpolation_input

        Returns
        -------
        + transform_lookup: dict
            A dictionary to lookup transformation results.
            The keys are the token in hash_index.
            The values are corresponding results.
        """
        hsh = np.unique(list(hash_index.values()))
        transform_lookup = {}
        if self.face is None:
            for hs in hsh:
                transform_lookup[hs] = None
            return transform_lookup
        for hs in hsh:
            main_key = get_key_by_value(hash_index, hs)
            var_name, dims, knw = main_dict[main_key]
            if isinstance(var_name, str):
                transform_lookup[hs] = None
            elif isinstance(var_name, tuple):
                uind, vind = index_lookup[hs]
                uind_dic = dict(zip(dims[0], uind))
                vind_dic = dict(zip(dims[1], vind))
                # This only matters when dim == 'face', no need to think about 'Xp1'
                (UfromUvel, UfromVvel, _, _) = self.ocedata.tp.four_matrix_for_uv(
                    uind_dic["face"][:, :, 0, 0]
                )

                (_, _, VfromUvel, VfromVvel) = self.ocedata.tp.four_matrix_for_uv(
                    vind_dic["face"][:, :, 0, 0]
                )

                UfromUvel = np.round(UfromUvel)
                UfromVvel = np.round(UfromVvel)
                VfromUvel = np.round(VfromUvel)
                VfromVvel = np.round(VfromVvel)

                bool_ufromu = np.abs(UfromUvel).astype(bool)
                bool_ufromv = np.abs(UfromVvel).astype(bool)
                bool_vfromu = np.abs(VfromUvel).astype(bool)
                bool_vfromv = np.abs(VfromVvel).astype(bool)

                indufromu = tuple(idid[bool_ufromu] for idid in uind)
                indufromv = tuple(idid[bool_ufromv] for idid in uind)
                indvfromu = tuple(idid[bool_vfromu] for idid in vind)
                indvfromv = tuple(idid[bool_vfromv] for idid in vind)

                transform_lookup[hs] = (
                    (UfromUvel, UfromVvel, VfromUvel, VfromVvel),
                    (bool_ufromu, bool_ufromv, bool_vfromu, bool_vfromv),
                    (indufromu, indufromv, indvfromu, indvfromv),
                )
            else:
                raise ValueError(f"unsupported dims: {dims}")
        # modify the index_lookup
        return transform_lookup

    def _mask_value_and_register(
        self, index_lookup, transform_lookup, hash_mask, hash_index, main_dict
    ):
        """Create masks for interpolation.

        Perform the masking process and register in a dictionary.

        Parameters
        ----------
        index_lookup: dict
            See _fatten_required_index_and_register
        transform_lookup: dict
            See _transform_vector_and_lookup
        hash_mask: dict
            See _register_interpolation_input
        hash_index: dict
            See _register_interpolation_input
        main_dict: dict
            See _register_interpolation_input

        Returns
        -------
        + mask_lookup: dict
            A dictionary to lookup masking results.
            The keys are the token in hash_mask.
            The values are corresponding results.
        """
        hsh = np.unique(list(hash_mask.values()))
        mask_lookup = {}
        for hs in hsh:
            main_key = get_key_by_value(hash_mask, hs)
            var_name, dims, knw = main_dict[main_key]
            hsind = hash_index[main_key]
            longDims = "".join([str(a_thing) for a_thing in dims])
            if isinstance(knw, KnW):
                ignore_mask = knw.ignore_mask
            elif isinstance(knw, tuple):
                ignore_mask = (knw[0].ignore_mask) or (knw[1].ignore_mask)

            if ignore_mask or ("X" not in longDims) or ("Y" not in longDims):
                mask_lookup[hs] = None
            elif isinstance(var_name, str):
                ind = index_lookup[hsind]
                ind_for_mask = _ind_for_mask(ind, dims)
                if "Zl" in dims:
                    cuvw = "Wvel"
                elif "Z" in dims:
                    if "Xp1" in dims and "Yp1" in dims:
                        raise NotImplementedError(
                            "The masking of corner points are open to "
                            "interpretations thus not implemented, "
                            "let knw.ignore_mask =True to go around"
                        )
                    elif "Xp1" in dims:
                        cuvw = "U"
                    elif "Yp1" in dims:
                        cuvw = "V"
                    else:
                        cuvw = "C"
                else:
                    cuvw = "C"
                masked = get_masked(self.ocedata, ind_for_mask, cuvwg=cuvw)
                mask_lookup[hs] = masked
            elif isinstance(var_name, tuple):
                to_unzip = transform_lookup[hsind]
                uind, vind = index_lookup[hsind]
                if to_unzip is None:
                    uind_for_mask = _ind_for_mask(uind, dims[0])
                    vind_for_mask = _ind_for_mask(vind, dims[1])
                    umask = get_masked(self.ocedata, uind_for_mask, cuvwg="U")
                    vmask = get_masked(self.ocedata, vind_for_mask, cuvwg="V")
                else:
                    (
                        _,
                        (bool_ufromu, bool_ufromv, bool_vfromu, bool_vfromv),
                        (indufromu, indufromv, indvfromu, indvfromv),
                    ) = to_unzip
                    umask = np.full(uind[0].shape, np.nan)
                    vmask = np.full(vind[0].shape, np.nan)

                    umask[bool_ufromu] = get_masked(
                        self.ocedata, _ind_for_mask(indufromu, dims[0]), cuvwg="U"
                    )
                    umask[bool_ufromv] = get_masked(
                        self.ocedata, _ind_for_mask(indufromv, dims[1]), cuvwg="V"
                    )
                    vmask[bool_vfromu] = get_masked(
                        self.ocedata, _ind_for_mask(indvfromu, dims[0]), cuvwg="U"
                    )
                    vmask[bool_vfromv] = get_masked(
                        self.ocedata, _ind_for_mask(indvfromv, dims[1]), cuvwg="V"
                    )
                mask_lookup[hs] = (umask, vmask)
        return mask_lookup

    def _read_data_and_register(
        self,
        index_lookup,
        transform_lookup,
        hash_read,
        hash_index,
        main_dict,
        prefetch_dict,
    ):
        """Read the data and register them as dict.

        Parameters
        ----------
        index_lookup: dict
            See _fatten_required_index_and_register
        transform_lookup: dict
            See _transform_vector_and_lookup
        hash_read: dict
            See _register_interpolation_input
        hash_index: dict
            See _register_interpolation_input
        main_dict: dict
            See _register_interpolation_input
        prefetch_dict: dict
            See _register_interpolation_input

        Returns
        -------
        + read_lookup: dict
            A dictionary to lookup data reading results.
            The keys are the token in hash_read.
            The values are corresponding results.
        """
        hsh = np.unique(list(hash_read.values()))
        data_lookup = {}
        for hs in hsh:
            main_key = get_key_by_value(hash_read, hs)
            hsind = hash_index[main_key]
            var_name, dims, knw = main_dict[main_key]
            prefetched, prefetch_prefix = prefetch_dict[main_key]
            if isinstance(var_name, str):
                ind = index_lookup[hsind]
                if prefetched is not None:
                    if prefetch_prefix is None:
                        raise ValueError(
                            "please pass value of the prefix of prefetched dataset, "
                            "even if the prefix is zero"
                        )
                    temp_ind = _subtract_prefetch_prefix(ind, prefetch_prefix)
                    needed = np.nan_to_num(prefetched[temp_ind])
                else:
                    needed = np.nan_to_num(smart_read(self.ocedata[var_name], ind))
                data_lookup[hs] = needed
            elif isinstance(var_name, tuple):
                uname, vname = var_name
                uind, vind = index_lookup[hsind]
                (
                    (UfromUvel, UfromVvel, VfromUvel, VfromVvel),
                    (bool_ufromu, bool_ufromv, bool_vfromu, bool_vfromv),
                    (indufromu, indufromv, indvfromu, indvfromv),
                ) = transform_lookup[hsind]
                temp_n_u = np.full(uind[0].shape, np.nan)
                temp_n_v = np.full(vind[0].shape, np.nan)
                if prefetched is not None:
                    upre, vpre = prefetched
                    ufromu = np.nan_to_num(
                        upre[_subtract_prefetch_prefix(indufromu, prefetch_prefix)]
                    )
                    ufromv = np.nan_to_num(
                        vpre[_subtract_prefetch_prefix(indufromv, prefetch_prefix)]
                    )
                    vfromu = np.nan_to_num(
                        upre[_subtract_prefetch_prefix(indvfromu, prefetch_prefix)]
                    )
                    vfromv = np.nan_to_num(
                        vpre[_subtract_prefetch_prefix(indvfromv, prefetch_prefix)]
                    )
                else:
                    ufromu = np.nan_to_num(smart_read(self.ocedata[uname], indufromu))
                    ufromv = np.nan_to_num(smart_read(self.ocedata[vname], indufromv))
                    vfromu = np.nan_to_num(smart_read(self.ocedata[uname], indvfromu))
                    vfromv = np.nan_to_num(smart_read(self.ocedata[vname], indvfromv))
                temp_n_u[bool_ufromu] = ufromu  # 0#
                temp_n_u[bool_ufromv] = ufromv  # 1#
                temp_n_v[bool_vfromu] = vfromu  # 0#
                temp_n_v[bool_vfromv] = vfromv  # 1#

                n_u = np.einsum("nijk,ni->nijk", temp_n_u, UfromUvel) + np.einsum(
                    "nijk,ni->nijk", temp_n_u, UfromVvel
                )
                n_v = np.einsum("nijk,ni->nijk", temp_n_v, VfromUvel) + np.einsum(
                    "nijk,ni->nijk", temp_n_v, VfromVvel
                )
                data_lookup[hs] = (n_u, n_v)

        return data_lookup

    def _compute_weight_and_register(
        self, mask_lookup, hash_weight, hash_mask, main_dict
    ):
        """Compute the weights and register them as dict.

        Parameters
        ----------
        mask_lookup: dict
            See _mask_value_and_register
        hash_weight: dict
            See _register_interpolation_input
        hash_mask: dict
            See _register_interpolation_input
        main_dict: dict
            See _register_interpolation_input

        Returns
        -------
        + weight_lookup: dict
            A dictionary to lookup the weights computed.
            The keys are the token in hash_weight.
            The values are corresponding results.
        """
        hsh = np.unique(list(hash_weight.values()))
        weight_lookup = {}
        for hs in hsh:
            main_key = get_key_by_value(hash_weight, hs)
            hsmsk = hash_mask[main_key]
            var_name, dims, knw = main_dict[main_key]
            masked = mask_lookup[hsmsk]
            if isinstance(var_name, tuple):
                ori_dims = dims
                dims = ori_dims[0]
                ori_knw = knw
                knw = ori_knw[0]
                # Assuming the two kernels have the same
                # vertical dimensions, which is reasonable.

            # shared part for 'vertical direction'
            this_bottom_scheme = "no_flux"
            if "Z" in dims:
                if self.rz is not None:
                    if knw.vkernel == "nearest":
                        rz = self.rz
                    else:
                        rz = self.rz_lin
                else:
                    rz = 0
            elif "Zl" in dims:
                this_bottom_scheme = None
                if self.rzl is not None:
                    if knw.vkernel == "nearest":
                        rz = self.rzl
                    else:
                        rz = self.rzl_lin
                else:
                    rz = 0
            else:
                rz = 0
            if self.rt is not None:
                if knw.tkernel == "nearest":
                    rt = self.rt
                else:
                    rt = self.rt_lin
            else:
                rt = 0

            if isinstance(var_name, str):
                if "Xp1" in dims:
                    rx = self.rx + 0.5
                else:
                    rx = self.rx
                if "Yp1" in dims:
                    ry = self.ry + 0.5
                else:
                    ry = self.ry
                if masked is None:
                    pk4d = None
                else:
                    pk4d = _find_pk_4d(masked, inheritance=knw.inheritance)
                weight = knw.get_weight(
                    rx=rx,
                    ry=ry,
                    rz=rz,
                    rt=rt,
                    pk4d=pk4d,
                    bottom_scheme=this_bottom_scheme,
                )
                weight_lookup[hs] = weight
            elif isinstance(var_name, tuple):
                uknw, vknw = ori_knw
                if masked is None:
                    upk4d = None
                    vpk4d = None
                else:
                    umask, vmask = masked
                    upk4d = _find_pk_4d(umask, inheritance=uknw.inheritance)
                    vpk4d = _find_pk_4d(vmask, inheritance=vknw.inheritance)
                uweight = uknw.get_weight(
                    self.rx + 1 / 2, self.ry, rz=rz, rt=rt, pk4d=upk4d
                )
                vweight = vknw.get_weight(
                    self.rx, self.ry + 1 / 2, rz=rz, rt=rt, pk4d=vpk4d
                )
                weight_lookup[hs] = (uweight, vweight)
        return weight_lookup

    def interpolate(
        self, var_name, knw, vec_transform=True, prefetched=None, prefetch_prefix=None
    ):
        """Do interpolation.

        This is the method that does the actual interpolation/derivative.
        It is a combination of the following methods:
        _register_interpolation_input,
        _fatten_required_index_and_register,
        _transform_vector_and_register,
        _read_data_and_register,
        _mask_value_and_register,
        _compute_weight_and_registe,.

        Parameters
        ----------
        var_name: list, str, tuple
            The variables to interpolate. Tuples are used for horizontal vectors.
            Put str and list in a list if you have multiple things to interpolate.
            This input also defines the format of the output.
        knw: KnW object, tuple, list, dict
            The kernel object used for the operation.
            Put them in the same order as var_name.
            Some level of automatic broadcasting is also supported.
        vec_transform: Boolean
            Whether to project the vector field to the local zonal/meridional direction.
        prefetched: numpy.ndarray, tuple, list, dict, None, default None
            The prefetched array for the data, this will effectively overwrite var_name.
            Put them in the same order as var_name.
            Some level of automatic broadcasting is also supported.
        prefetch_prefix: tuple, list, dict, None, default None
            The prefix of the prefetched array.
            Put them in the same order as var_name.
            Some level of automatic broadcasting is also supported.

        Returns
        -------
        to_return: list, numpy.array, tuple
            The interpolation/derivative output in the same format as var_name.
        """
        to_return = []
        (
            output_format,
            main_keys,
            prefetch_dict,
            main_dict,
            hash_index,
            hash_mask,
            hash_read,
            hash_weight,
        ) = self._register_interpolation_input(
            var_name, knw, prefetched=prefetched, prefetch_prefix=prefetch_prefix
        )
        index_lookup = self._fatten_required_index_and_register(hash_index, main_dict)
        transform_lookup = self._transform_vector_and_register(
            index_lookup, hash_index, main_dict
        )
        data_lookup = self._read_data_and_register(
            index_lookup,
            transform_lookup,
            hash_read,
            hash_index,
            main_dict,
            prefetch_dict,
        )
        mask_lookup = self._mask_value_and_register(
            index_lookup, transform_lookup, hash_mask, hash_index, main_dict
        )
        weight_lookup = self._compute_weight_and_register(
            mask_lookup, hash_weight, hash_mask, main_dict
        )
        # index_list = []
        for key in main_keys:
            var_name, dims, knw = main_dict[key]
            if isinstance(var_name, str):
                needed = data_lookup[hash_read[key]]
                weight = weight_lookup[hash_weight[key]]
                needed = _partial_flatten(needed)
                weight = _partial_flatten(weight)
                to_return.append(np.einsum("nj,nj->n", needed, weight))
                # index_list.append((index_lookup[hash_index[key]],
                #                    transform_lookup[hash_index[key]],
                #                    data_lookup[hash_read[key]]))
            elif isinstance(var_name, tuple):
                n_u, n_v = data_lookup[hash_read[key]]
                uweight, vweight = weight_lookup[hash_weight[key]]
                u = np.einsum("nijk,nijk->n", n_u, uweight)
                v = np.einsum("nijk,nijk->n", n_v, vweight)
                if vec_transform:
                    u, v = local_to_latlon(u, v, self.cs, self.sn)
                to_return.append((u, v))
                # index_list.append((index_lookup[hash_index[key]],
                #                    transform_lookup[hash_index[key]],
                #                    data_lookup[hash_read[key]]))
            else:
                raise ValueError(f"unexpected var_name: {var_name}")

        final_dict = dict(zip(output_format["final_var_name"], to_return))
        ori_list = output_format["ori_list"]
        output = []
        # print(ori_list,to_return,final_dict.keys())
        for key in ori_list:
            var, knw = key

            if var is None:
                output.append(None)
            elif isinstance(var, tuple):
                if self.face is None:
                    temp_key = [(var[i], knw[i]) for i in range(len(var))]
                    values = tuple(final_dict.get(k) for k in temp_key)
                    output.append(values)
                else:
                    output.append(final_dict.get(key))
            else:
                output.append(final_dict.get(key))
        if output_format["single"]:
            output = output[0]
        return output



import warnings

import numpy as np

from seaduck.eulerian import Position
from seaduck.kernel_weight import KnW
from seaduck.lagrangian import Particle, uknw, vknw
from seaduck.ocedata import OceData
from seaduck.utils import convert_time

lagrange_token = "__particle."


def OceInterp(
    od,
    var_list,
    x,
    y,
    z,
    t,
    kernel_list=None,
    lagrangian=False,
    lagrange_kwarg={},
    update_stops="default",
    return_in_between=True,
    return_pt_time=True,
    kernel_kwarg={},
):
    """Interp for people who just want to take a quick look.

    **This is the centerpiece function of the package, through which
    you can access almost all of its functionality.**.

    Parameters
    ----------
    od: OceInterp.OceData object or xarray.Dataset (limited support for netCDF Dataset)
        The dataset to work on.
    var_list: str or list
        A list of variable or pair of variables.
    kernel_list: OceInterp.KnW or list of OceInterp.KnW objects, optional
        Indicates which kernel to use for each interpolation.
    x, y, z: numpy.ndarray, float
        The location of the particles, where x and y are in degrees,
        and z is in meters (deeper locations are represented by more negative values).
    t: numpy.ndarray, float, string/numpy.datetime64
        In the Eulerian scheme, this represents the time of interpolation.
        In the Lagrangian scheme, it represents the time needed for output.
    lagrangian: bool, default False
        Specifies whether the interpolation is done in the Eulerian or Lagrangian scheme.
    lagrange_kwarg: dict, optional
        Keyword arguments passed into the OceInterp.lagrangian.Particle object.
    update_stops: None, 'default', or iterable of float
        Specifies the time to update the prefetch velocity.
    return_in_between: bool, default True
        In Lagrangian mode, this returns the interpolation not only at time t,
        but also at every point in time when the speed is updated.
    return_pt_time: bool, default True
        Specifies whether to return the time of all the steps.
    kernel_kwarg: dict, optional
        keyword arguments to pass into seaduck.KnW object.
    """
    if not isinstance(od, OceData):
        od = OceData(od)

    if isinstance(var_list, dict):
        kernel_list = list(var_list.values())
        var_list = list(var_list.keys())
    elif isinstance(var_list, str):
        var_list = [var_list]
    elif isinstance(var_list, tuple):
        var_list = [var_list]
    elif isinstance(var_list, list):
        pass
    else:
        raise ValueError("var_list type not recognized.")
    if isinstance(kernel_list, list):
        pass
    elif kernel_list is None:
        kernel_list = []
        the_kernel = KnW(**kernel_kwarg)
        for i in var_list:
            if isinstance(i, str):
                kernel_list.append(the_kernel)
            elif isinstance(i, tuple):
                if kernel_kwarg != {}:
                    kernel_list.append((the_kernel, the_kernel))
                else:
                    kernel_list.append((uknw, vknw))
            else:
                raise ValueError(
                    "members of var_list need to be made up of string or tuples"
                )
    if isinstance(t, np.ndarray):
        if np.issubdtype(t.dtype, np.datetime64) or np.issubdtype(t.dtype, str):
            t = np.array([convert_time(some_t) for some_t in t])
    elif isinstance(t, (np.datetime64, str)):
        t = convert_time(t)
    elif isinstance(t, float):
        pass
    else:
        raise ValueError("time format not supported")
    if not lagrangian:
        pt = Position()
        pt.from_latlon(x=x, y=y, z=z, t=t, data=od)
        for i, var in enumerate(var_list):
            if lagrange_token in var:
                raise AttributeError(
                    "__particle variables is only available for Lagrangian Particles"
                )
        to_return = pt.interpolate(var_list, kernel_list)
        return to_return

    else:
        try:
            assert len(t) > 1
        except AssertionError as exc:
            raise ValueError(
                "There needs to be at least two time steps to run the lagrangian Particle"
            ) from exc
        t_start = t[0]
        t_nec = t[1:]
        pt = Particle(
            x=x, y=y, z=z, t=np.ones_like(x) * t_start, data=od, **lagrange_kwarg
        )
        stops, raw = pt.to_list_of_time(
            t_nec, update_stops=update_stops, return_in_between=return_in_between
        )
        to_return = []
        for i, var in enumerate(var_list):
            if var == lagrange_token + "raw":
                to_return.append(raw)
            elif lagrange_token in var:
                sublist = []
                for snap in raw:
                    sublist.append(getattr(snap, var[len(lagrange_token) :]))
                to_return.append(sublist)
            else:
                sublist = []
                for snap in raw:
                    sublist.append(snap.interpolate(var, kernel_list[i]))
                to_return.append(sublist)

        if return_pt_time:
            return stops, to_return
        else:
            if return_in_between:
                warnings.warn("Some of the returns is not on the times you specified.")
            return to_return



import numpy as np
import pytest
import xarray as xr

import seaduck as sd
from seaduck import utils


@pytest.fixture
def incomplete_data(request):
    ds = utils.get_dataset("curv")
    if request.param == "drop_YG":
        od = sd.OceData(ds)
        ds_out = od._ds.drop_vars(["YG"])
        od._add_missing_cs_sn()
        ds_out["CS"] = xr.DataArray(od["CS"], dims=od._ds["XC"].dims)
        ds_out["SN"] = xr.DataArray(od["SN"], dims=od._ds["XC"].dims)
        return ds_out
    elif request.param == "drop_dyG":
        return ds.drop_vars(["dyG"])
    elif request.param == "drop_drF":
        return ds.drop_vars(["drF"])
    elif request.param == "drop_drF_and_Zp1":
        return ds.drop_vars(["drF", "Zp1"])
    elif request.param == "drop_time_midp":
        return ds.drop_vars(
            ["time_midp"] + [i for i in ds.data_vars if "time_midp" in ds[i].dims]
        )


@pytest.mark.parametrize(
    "incomplete_data",
    ["drop_YG", "drop_dyG", "drop_time_midp", "drop_drF", "drop_drF_and_Zp1"],
    indirect=True,
)
def test_incomplete_data(incomplete_data):
    oo = sd.OceData(incomplete_data)
    hrel = oo._find_rel_h(np.array([-14]), np.array([70.5]))
    assert isinstance(hrel, sd.ocedata.HRel)


@pytest.mark.parametrize("ds", ["curv"], indirect=True)
def test_auto_alias(ds):
    with pytest.raises(NotImplementedError):
        sd.OceData(ds, alias="auto")


@pytest.mark.parametrize("ds", ["curv"], indirect=True)
def test_manual_alias(ds):
    an_alias = {
        "dXC": "dxC",
        "dYC": "dyC",
        "dZ": "drC",
        "dXG": "dxG",
        "dYG": "dyG",
        "dZl": "drF",
        "SALT": "S",
    }
    od = sd.OceData(ds, an_alias)
    od["SALT"] = xr.ones_like(od._ds["XC"])
    try:
        od.show_alias()
    except ImportError:
        with pytest.raises(NameError):
            od.show_alias()


@pytest.mark.parametrize("ds", ["curv"], indirect=True)
def test_add_missing_grid(ds):
    od = sd.OceData(ds)
    od._add_missing_cs_sn()
    assert isinstance(od["CS"], np.ndarray)
    # TODO
    od._add_missing_grid()


@pytest.mark.parametrize("ds", ["curv"], indirect=True)
def test_not_h_ready(ds):
    temp = ds.drop_vars(["XG", "YG"])
    with pytest.raises(ValueError):
        sd.OceData(temp)


@pytest.mark.parametrize("incomplete_data", ["drop_YG"], indirect=True)
@pytest.mark.parametrize("lat", list(np.linspace(70, 70.618, 3)))
@pytest.mark.parametrize("lon", list(np.linspace(-17, -13.618, 3)))
def test_accurate_reproduce_local_cartesian(incomplete_data, lat, lon):
    tub = sd.OceData(incomplete_data)
    lon = np.array([lon])
    lat = np.array([lat])
    hrel = tub._find_rel_h(lon, lat)
    assert isinstance(hrel, sd.ocedata.HRel)
    assert "=" in repr(hrel)
    face, iy, ix, rx, ry, cs, sn, dx, dy, bx, by = hrel.values()
    new_lon, new_lat = sd.utils.rel2latlon(rx, ry, cs, sn, dx, dy, bx, by)
    assert np.allclose(new_lon, lon)
    assert np.allclose(new_lat, lat)


@pytest.mark.parametrize("od", ["aviso"], indirect=True)
@pytest.mark.parametrize("lat", list(np.linspace(-74.618, -44.0, 3)))
@pytest.mark.parametrize("lon", list(np.linspace(-179.618, 179.618, 4)))
def test_accurate_reproduce_rectilinear(od, lat, lon):
    lon = np.array([lon])
    lat = np.array([lat])
    hrel = od._find_rel_h(lon, lat)
    assert isinstance(hrel, sd.ocedata.HRel)
    face, iy, ix, rx, ry, cs, sn, dx, dy, bx, by = hrel.values()
    new_lon, new_lat = sd.utils.rel2latlon(rx, ry, cs, sn, dx, dy, bx, by)
    assert np.allclose(new_lon, lon)
    assert np.allclose(new_lat, lat)


@pytest.mark.parametrize("od", ["curv"], indirect=True)
@pytest.mark.parametrize("style", ["nearest", "linear"])
@pytest.mark.parametrize("t", [1.85e9, 2e9])
def test_accurate_reproduce_time(od, style, t):
    t = np.array([t])
    if style == "nearest":
        trel = od._find_rel_t(t)
        assert isinstance(trel, sd.ocedata.TRel)
    elif style == "linear":
        trel = od._find_rel_t_lin(t)
        assert isinstance(trel, sd.ocedata.TLinRel)
    it, rt, dt, bt = trel.values()
    assert np.allclose(bt + dt * rt, t)


@pytest.mark.parametrize("od", ["curv"], indirect=True)
@pytest.mark.parametrize("style", ["nearest", "linear"])
@pytest.mark.parametrize("z", [0, -5, -35])
def test_accurate_reproduce_z(od, style, z):
    z = np.array([z])
    if style == "nearest":
        vrel = od._find_rel_v(z)
        assert isinstance(vrel, sd.ocedata.VRel)
    elif style == "linear":
        vrel = od._find_rel_v_lin(z)
        assert isinstance(vrel, sd.ocedata.VLinRel)
    iz, rz, dz, bz = vrel.values()
    assert np.allclose(bz + dz * rz, z)


@pytest.mark.parametrize("od", ["curv"], indirect=True)
@pytest.mark.parametrize("style", ["nearest", "linear"])
@pytest.mark.parametrize("z", [0, -5, -35])
def test_accurate_reproduce_z_staggered(od, style, z):
    z = np.array([z])
    if style == "nearest":
        vrel = od._find_rel_vl(z)
        assert isinstance(vrel, sd.ocedata.VlRel)
    elif style == "linear":
        vrel = od._find_rel_vl_lin(z)
        assert isinstance(vrel, sd.ocedata.VlLinRel)
    iz, rz, dz, bz = vrel.values()
    assert np.allclose(bz + dz * rz, z)



import numpy as np

try:  # pragma: no cover
    import xgcm
except ImportError:  # pragma: no cover
    pass


def _raise_if_no_xgcm():
    try:
        import xgcm

        xgcm
    except ImportError:
        raise ImportError(
            "The python package xgcm is needed."
            "You can install it with:"
            "conda install -c xgcm"
        )


def create_ecco_grid(ds, for_outer=False):
    _raise_if_no_xgcm()  # pragma: no cover
    face_connections = {
        "face": {
            0: {"X": ((12, "Y", False), (3, "X", False)), "Y": (None, (1, "Y", False))},
            1: {
                "X": ((11, "Y", False), (4, "X", False)),
                "Y": ((0, "Y", False), (2, "Y", False)),
            },
            2: {
                "X": ((10, "Y", False), (5, "X", False)),
                "Y": ((1, "Y", False), (6, "X", False)),
            },
            3: {"X": ((0, "X", False), (9, "Y", False)), "Y": (None, (4, "Y", False))},
            4: {
                "X": ((1, "X", False), (8, "Y", False)),
                "Y": ((3, "Y", False), (5, "Y", False)),
            },
            5: {
                "X": ((2, "X", False), (7, "Y", False)),
                "Y": ((4, "Y", False), (6, "Y", False)),
            },
            6: {
                "X": ((2, "Y", False), (7, "X", False)),
                "Y": ((5, "Y", False), (10, "X", False)),
            },
            7: {
                "X": ((6, "X", False), (8, "X", False)),
                "Y": ((5, "X", False), (10, "Y", False)),
            },
            8: {
                "X": ((7, "X", False), (9, "X", False)),
                "Y": ((4, "X", False), (11, "Y", False)),
            },
            9: {"X": ((8, "X", False), None), "Y": ((3, "X", False), (12, "Y", False))},
            10: {
                "X": ((6, "Y", False), (11, "X", False)),
                "Y": ((7, "Y", False), (2, "X", False)),
            },
            11: {
                "X": ((10, "X", False), (12, "X", False)),
                "Y": ((8, "Y", False), (1, "X", False)),
            },
            12: {
                "X": ((11, "X", False), None),
                "Y": ((9, "Y", False), (0, "X", False)),
            },
        }
    }
    coords = {
        "X": {"center": "X", "left": "Xp1"},
        "Y": {"center": "Y", "left": "Yp1"},
        "Z": {"center": "Z", "left": "Zl"},
        "time": {"center": "time", "inner": "time_midp"},
    }
    if for_outer:
        coords["Z"] = {"center": "Z", "outer": "Zl"}
    xgcmgrd = xgcm.Grid(
        ds, periodic=False, face_connections=face_connections, coords=coords
    )
    return xgcmgrd


def create_periodic_grid(ds):
    _raise_if_no_xgcm()  # pragma: no cover
    xgcmgrd = xgcm.Grid(
        ds,
        periodic=["X"],
        coords={
            "X": {"center": "X", "outer": "Xp1"},
            "Y": {"center": "Y", "outer": "Yp1"},
            "Z": {"center": "Z", "left": "Zl"},
            "time": {"center": "time", "outer": "time_outer"},
        },
    )
    return xgcmgrd


def hor_div(tub, xgcmgrd, xfluxname, yfluxname):
    """Calculate horizontal divergence using xgcm.

    Parameters
    ----------
    tub: sd.OceData or xr.Dataset
        The dataset to calculate data from
    xgcmgrd: xgcm.Grid
        The Grid of the dataset
    xfluxname, yfluxname: string
        The name of the variables corresponding to the horizontal fluxes
        in concentration m^3/s
    """
    try:
        tub["Vol"]
    except KeyError:
        tub._add_missing_vol()
    xy_diff = xgcmgrd.diff_2d_vector(
        {"X": tub[xfluxname].fillna(0), "Y": tub[yfluxname].fillna(0)},
        boundary="fill",
        fill_value=0.0,
    )
    x_diff = xy_diff["X"]
    y_diff = xy_diff["Y"]
    hConv = -(x_diff + y_diff) / tub["Vol"]
    return hConv


def ver_div(tub, xgcmgrd, zfluxname):
    """Calculate horizontal divergence using xgcm.

    Parameters
    ----------
    tub: sd.OceData or xr.Dataset
        The dataset to calculate data from
    xgcmgrd: xgcm.Grid
        The Grid of the dataset
    xfluxname, yfluxname, zfluxname: string
        The name of the variables corresponding to the fluxes
        in concentration m^3/s
    """
    try:
        tub["Vol"]
    except KeyError:
        tub._add_missing_vol()
    vConv = (
        xgcmgrd.diff(tub[zfluxname].fillna(0), "Z", boundary="fill", fill_value=0.0)
        / tub["Vol"]
    )
    return vConv


def total_div(tub, xgcmgrd, xfluxname, yfluxname, zfluxname):
    """Calculate 3D divergence using xgcm.

    Parameters
    ----------
    tub: sd.OceData or xr.Dataset
        The dataset to calculate data from
    xgcmgrd: xgcm.Grid
        The Grid of the dataset
    zfluxname: string
        The name of the variables corresponding to the vertical flux
        in concentration m^3/s
    """
    hConv = hor_div(tub, xgcmgrd, xfluxname, yfluxname)
    vConv = ver_div(tub, xgcmgrd, zfluxname)
    return hConv + vConv


def bolus_vel_from_psi(tub, xgcmgrd, psixname="GM_PsiX", psiyname="GM_PsiY"):
    strmx = tub[psixname].fillna(0)
    strmy = tub[psiyname].fillna(0)

    u = xgcmgrd.diff(strmx, "Z", boundary="fill", fill_value=0.0) / tub["drF"]
    v = xgcmgrd.diff(strmy, "Z", boundary="fill", fill_value=0.0) / tub["drF"]

    vstrmx = strmx * tub["dyG"]
    vstrmy = strmy * tub["dxG"]

    xy_diff = xgcmgrd.diff_2d_vector(
        {"X": vstrmx, "Y": vstrmy}, boundary="fill", fill_value=0.0
    )
    x_diff = xy_diff["X"]
    y_diff = xy_diff["Y"]
    hDiv = x_diff + y_diff

    w = hDiv / tub["rA"]
    return u, v, w


def _slice_corner(array, fc, iy1, iy2, ix1, ix2):
    left = np.minimum(ix1, ix2)
    righ = np.maximum(ix1, ix2)
    down = np.minimum(iy1, iy2)
    uppp = np.maximum(iy1, iy2)
    return array[..., fc, down : uppp + 1, left : righ + 1]


def _right90(array):
    return array[..., ::-1].transpose([0, 2, 1])


def _left90(array):
    return array[..., ::-1, :].transpose([0, 2, 1])


def buffer_x_withface(s, face, lm, rm, tp):
    """Create buffer zone in x direction for one face of an array.

    Parameters
    ----------
    s: numpy.ndarray
        the center field, the last dimension being X,
        the second dimension being face.
    face: int
        which face to create buffer for.
    lm: int
        the size of the margin to the left.
    rm: int
        the size of the margin to the right.
    tp: seaduck.Topology
        the topology object of the
    """
    shape = list(s.shape)
    shape.pop(1)
    shape[-1] += lm + rm
    xbuffer = np.zeros(shape)
    xbuffer[..., lm:-rm] = s[:, face]
    try:
        fc1, iy1, ix1 = tp.ind_moves((face, tp.iymax, 0), [2 for i in range(lm)])
        fc2, iy2, ix2 = tp.ind_moves((face, 0, 0), [2])
        left = _slice_corner(s, fc1, iy1, iy2, ix1, ix2)
    except IndexError:
        left = np.zeros_like(xbuffer[..., :lm])

    try:
        fc3, iy3, ix3 = tp.ind_moves((face, tp.iymax, tp.ixmax), [3 for i in range(rm)])
        fc4, iy4, ix4 = tp.ind_moves((face, 0, tp.ixmax), [3])
        righ = _slice_corner(s, fc3, iy3, iy4, ix3, ix4)
    except IndexError:
        righ = np.zeros_like(xbuffer[..., -rm:])
    try:
        xbuffer[..., :lm] = left
    except ValueError:
        xbuffer[..., :lm] = _right90(left)

    try:
        xbuffer[..., -rm:] = righ
    except ValueError:
        xbuffer[..., -rm:] = _right90(righ)

    return xbuffer


def buffer_y_withface(s, face, lm, rm, tp):
    """Create buffer zone in x direction for one face of an array.

    Parameters
    ----------
    s: numpy.ndarray
        the center field, the last dimension being X,
        the second dimension being face.
    face: int
        which face to create buffer for.
    lm: int
        the size of the margin to the bottom.
    rm: int
        the size of the margin to the top.
    tp: seaduck.Topology
        the topology object of the
    """
    shape = list(s.shape)
    shape.pop(1)
    shape[-2] += lm + rm
    ybuffer = np.zeros(shape)
    ybuffer[..., lm:-rm, :] = s[:, face]
    try:
        fc1, iy1, ix1 = tp.ind_moves((face, 0, tp.ixmax), [1 for i in range(lm)])
        fc2, iy2, ix2 = tp.ind_moves((face, 0, 0), [1])
        left = _slice_corner(s, fc1, iy1, iy2, ix1, ix2)
    except IndexError:
        left = np.zeros_like(ybuffer[..., :lm, :])

    try:
        fc3, iy3, ix3 = tp.ind_moves((face, tp.iymax, tp.ixmax), [0 for i in range(rm)])
        fc4, iy4, ix4 = tp.ind_moves((face, tp.iymax, 0), [0])
        righ = _slice_corner(s, fc3, iy3, iy4, ix3, ix4)
    except IndexError:
        righ = np.zeros_like(ybuffer[..., -rm:, :])
    try:
        ybuffer[..., :lm, :] = left
    except ValueError:
        ybuffer[..., :lm, :] = _left90(left)

    try:
        ybuffer[..., -rm:, :] = righ
    except ValueError:
        ybuffer[..., -rm:, :] = _left90(righ)
    return ybuffer


def buffer_x_periodic(s, lm, rm):
    shape = list(s.shape)
    shape[-1] += lm + rm
    xbuffer = np.zeros(shape)
    xbuffer[..., lm : shape[-1] - rm] = s
    if lm > 0:
        xbuffer[..., :lm] = s[..., -lm:]
    if rm > 0:
        xbuffer[..., -rm:] = s[..., :rm]
    return xbuffer


def buffer_y_periodic(s, lm, rm):
    shape = list(s.shape)
    shape[-2] += lm + rm
    ybuffer = np.zeros(shape)
    ybuffer[..., lm : shape[-2] - rm, :] = s
    if lm > 0:
        ybuffer[..., :lm, :] = s[..., -lm:, :]
    if rm > 0:
        ybuffer[..., -rm:, :] = s[..., :rm, :]
    return ybuffer


def buffer_z_nearest_withoutface(s, lm, rm):
    shape = list(s.shape)
    shape[-3] += lm + rm
    zbuffer = np.zeros(shape)
    zbuffer[..., lm : shape[-3] - rm, :, :] = s
    if lm > 0:
        zbuffer[..., :lm, :, :] = s[..., :1, :, :]
    if rm > 0:
        zbuffer[..., -rm:, :, :] = s[..., -1:, :, :]
    return zbuffer


def _slope_ratio(Rjm, Rj, Rjp, u_sign, not_z=1):
    """Calculate slope ratio for flux limiter."""
    cr_max = 1e6  # doesn't matter
    cr = np.zeros_like(u_sign)
    pos = not_z * u_sign > 0
    neg = not_z * u_sign <= 0
    cr[pos] = Rjm[pos]
    cr[neg] = Rjp[neg]
    zero_divide = np.abs(Rj) * cr_max <= np.abs(cr)
    cr[zero_divide] = np.sign(cr[zero_divide]) * np.sign(u_sign[zero_divide]) * cr_max
    cr[~zero_divide] = cr[~zero_divide] / Rj[~zero_divide]
    return cr


def superbee_fluxlimiter(cr):
    return np.maximum(0.0, np.maximum(np.minimum(1.0, 2 * cr), np.minimum(2.0, cr)))


def second_order_flux_limiter_x(s_center, u_cfl):
    xbuffer = buffer_x_periodic(s_center, 2, 2)
    deltas = np.nan_to_num(np.diff(xbuffer, axis=-1), 0)
    Rjp = deltas[..., 2:]
    Rj = deltas[..., 1:-1]
    Rjm = deltas[..., :-2]

    cr = _slope_ratio(Rjm, Rj, Rjp, u_cfl)
    limiter = superbee_fluxlimiter(cr)
    swall = (
        np.nan_to_num(xbuffer[..., 1:-2] + xbuffer[..., 2:-1]) * 0.5
        - np.sign(u_cfl) * ((1 - limiter) + u_cfl * limiter) * Rj * 0.5
    )
    return swall


def second_order_flux_limiter_y(s_center, u_cfl):
    ybuffer = buffer_y_periodic(s_center, 2, 2)
    deltas = np.nan_to_num(np.diff(ybuffer, axis=-2), 0)
    Rjp = deltas[..., 2:, :]
    Rj = deltas[..., 1:-1, :]
    Rjm = deltas[..., :-2, :]

    cr = _slope_ratio(Rjm, Rj, Rjp, u_cfl)
    limiter = superbee_fluxlimiter(cr)
    swall = (
        np.nan_to_num(ybuffer[..., 1:-2, :] + ybuffer[..., 2:-1, :]) * 0.5
        - np.sign(u_cfl) * ((1 - limiter) + u_cfl * limiter) * Rj * 0.5
    )
    return swall


def second_order_flux_limiter_z_withoutface(s_center, u_cfl):
    zbuffer = buffer_z_nearest_withoutface(s_center, 2, 1)
    deltas = np.nan_to_num(np.diff(zbuffer, axis=-3), 0)
    Rjp = deltas[..., 2:, :, :]
    Rj = deltas[..., 1:-1, :, :]
    Rjm = deltas[..., :-2, :, :]

    cr = _slope_ratio(Rjm, Rj, Rjp, u_cfl, not_z=-1)
    limiter = superbee_fluxlimiter(cr)
    # swall = np.nan_to_num(zbuffer[...,1:-2,:,:]+zbuffer[...,2:-1,:,:])*0.5- np.sign(u_cfl)**Rj*0.5
    swall = (
        np.nan_to_num(zbuffer[..., 1:-2, :, :] + zbuffer[..., 2:-1, :, :]) * 0.5
        + np.sign(u_cfl) * ((1 - limiter) + u_cfl * limiter) * Rj * 0.5
    )
    return swall


def third_order_upwind_z(s, w):
    """Get interpolated salinity in the vertical.

    for more info, see
    https://mitgcm.readthedocs.io/en/latest/algorithm/adv-schemes.html#third-order-upwind-bias-advection
    This function currently only work when there is no through surface flux.

    Parameters
    ----------
    s: numpy.ndarray
        the tracer field, the first dimension being Z
    w: numpy.ndarray
        the vertical velocity field, the first dimension being Zl
    """
    w[0] = 0
    deltas = np.nan_to_num(s[1:] - s[:-1], 0)
    Rj = np.zeros_like(s)
    Rj[1:] = deltas
    Rjp = np.roll(Rj, -1, axis=0)
    Rjm = np.roll(Rj, 1, axis=0)
    Rjjp = Rjp - Rj
    Rjjm = Rj - Rjm

    ssum = np.zeros_like(s)
    ssum[1:] = s[1:] + s[:-1]
    sz = 0.5 * (ssum - 1 / 6 * (Rjjp + Rjjm) - np.sign(w) * 1 / 6 * (Rjjp - Rjjm))
    return sz


def third_order_DST_x(xbuffer, u_cfl):
    """Get interpolated salinity in X.

    for more info, see
    https://mitgcm.readthedocs.io/en/latest/algorithm/adv-schemes.html#third-order-direct-space-time

    Parameters
    ----------
    xbuffer: numpy.ndarray
        the tracer field with buffer zone added (lm=2,rm=1), the last dimension being X
    u_cfl: numpy.ndarray
        the u velocity normalized by grid size in x, in s^-1.
    """
    lm = 2
    deltas = np.nan_to_num(np.diff(xbuffer, axis=-1), 0)
    Rjp = deltas[..., lm:]
    Rj = deltas[..., lm - 1 : -1]
    Rjm = deltas[..., lm - 2 : -2]

    d0 = (2.0 - u_cfl) * (1.0 - u_cfl) / 6
    d1 = (1.0 - u_cfl * u_cfl) / 6

    sx = 0.5 * (
        (1 + np.sign(u_cfl)) * (xbuffer[..., lm - 1 : -2] + (d0 * Rj + d1 * Rjm))
        + (1 - np.sign(u_cfl)) * (xbuffer[..., lm:-1] - (d0 * Rj + d1 * Rjp))
    )
    return sx


def third_order_DST_y(ybuffer, u_cfl):
    """Get interpolated salinity in Y.

    for more info, see
    https://mitgcm.readthedocs.io/en/latest/algorithm/adv-schemes.html#third-order-direct-space-time

    Parameters
    ----------
    ybuffer: numpy.ndarray
        the tracer field with buffer zone added (lm=2,rm=1), the second last dimension being Y
    u_cfl: numpy.ndarray
        the v velocity normalized by grid size in y, in s^-1.
    """
    lm = 2
    deltas = np.nan_to_num(np.diff(ybuffer, axis=-2), 0)
    Rjp = deltas[..., lm:, :]
    Rj = deltas[..., lm - 1 : -1, :]
    Rjm = deltas[..., lm - 2 : -2, :]

    d0 = (2.0 - u_cfl) * (1.0 - u_cfl) / 6
    d1 = (1.0 - u_cfl * u_cfl) / 6

    sy = 0.5 * (
        (1 + np.sign(u_cfl)) * (ybuffer[..., lm - 1 : -2, :] + (d0 * Rj + d1 * Rjm))
        + (1 - np.sign(u_cfl)) * (ybuffer[..., lm:-1, :] - (d0 * Rj + d1 * Rjp))
    )
    return sy



# Copyright 2023, Wenrui Jiang.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    # NOTE: the `version.py` file must not be present in the git repository
    #   as it is generated by setuptools at install time
    from seaduck.version import __version__
except ImportError:  # pragma: no cover
    # Local copy or not installed with setuptools
    __version__ = "999"

from seaduck.eulerian import Position
from seaduck.kernel_weight import KnW
from seaduck.lagrangian import Particle
from seaduck.ocedata import OceData
from seaduck.oceinterp import OceInterp
from seaduck.runtime_conf import rcParam
from seaduck.topology import Topology

__all__ = [
    "__version__",
    "Position",
    "KnW",
    "Particle",
    "OceData",
    "OceInterp",
    "rcParam",
    "Topology",
]



import numpy as np
import pytest

from seaduck.eulerian_budget import (
    buffer_x_periodic,
    buffer_y_periodic,
    buffer_z_nearest_withoutface,
    second_order_flux_limiter_x,
    second_order_flux_limiter_y,
    second_order_flux_limiter_z_withoutface,
    superbee_fluxlimiter,
)


@pytest.fixture
def random_4d():
    np.random.seed(401)
    return np.random.random((3, 4, 5, 4))


def test_superbee():
    cr = np.array([-1, 0.25, 0.5, 1, 2, 100])
    res = superbee_fluxlimiter(cr)
    assert np.allclose(res, np.array([0.0, 0.5, 1.0, 1.0, 2.0, 2.0]))


@pytest.mark.parametrize(["lm", "rm"], [(0, 2), (2, 1), (1, 0)])
def test_buffer_x_periodic(random_4d, lm, rm):
    buffer = buffer_x_periodic(random_4d, lm, rm)
    if rm != 0:
        assert np.allclose(buffer[..., -1], random_4d[..., rm - 1])
    if lm != 0:
        assert np.allclose(buffer[..., 0], random_4d[..., -lm])


@pytest.mark.parametrize(["lm", "rm"], [(0, 2), (2, 1), (1, 0)])
def test_buffer_y_periodic(random_4d, lm, rm):
    buffer = buffer_y_periodic(random_4d, lm, rm)
    if rm != 0:
        assert np.allclose(buffer[..., -1, :], random_4d[..., rm - 1, :])
    if lm != 0:
        assert np.allclose(buffer[..., 0, :], random_4d[..., -lm, :])


@pytest.mark.parametrize(["lm", "rm"], [(0, 2), (2, 1), (1, 0)])
def test_buffer_z_nearest_withoutface(random_4d, lm, rm):
    buffer = buffer_z_nearest_withoutface(random_4d, lm, rm)
    if rm != 0:
        assert np.allclose(buffer[..., -1, :, :], random_4d[..., -1, :, :])
    if lm != 0:
        assert np.allclose(buffer[..., 0, :, :], random_4d[..., 0, :, :])


# The next three tests are not very quantitative,
# but they has been tested with real datasets,
# still there needs to be some improvements.
def test_second_order_flux_limiter_x(random_4d):
    np.random.seed(401)
    u_cfl = np.random.random((3, 4, 5, 5)) * 2 - 1
    ans = second_order_flux_limiter_x(random_4d, u_cfl)
    assert ans.shape == u_cfl.shape
    assert ans.dtype == "float64"


def test_second_order_flux_limiter_y(random_4d):
    np.random.seed(401)
    v_cfl = np.random.random((3, 4, 6, 4)) * 2 - 1
    ans = second_order_flux_limiter_y(random_4d, v_cfl)
    assert ans.shape == v_cfl.shape
    assert ans.dtype == "float64"


def test_second_order_flux_limiter_z_withoutface(random_4d):
    np.random.seed(401)
    w_cfl = np.random.random((3, 4, 5, 4)) * 2 - 1
    ans = second_order_flux_limiter_z_withoutface(random_4d, w_cfl)
    assert ans.shape == w_cfl.shape
    assert ans.dtype == "float64"



import numpy as np
import pytest

import seaduck as sd

N = int(1e2)

# Change the vertical depth of the particles here
levels = np.array([-5])
sqrtN = int(np.sqrt(N))

# Change the longitude and latitude positions of the particles here
xx = np.linspace(-19, -9, sqrtN)
yy = np.linspace(63, 57, sqrtN)

# Compute intermediate grid variables
xxx, yyy = np.meshgrid(xx, yy)
x = xxx.ravel()
y = yyy.ravel()
x, z = np.meshgrid(x, levels)
y, z = np.meshgrid(y, levels)
x = x.ravel()
y = y.ravel()
z = z.ravel()

# Change the times here
start_time = "1992-01-17"
t = sd.utils.convert_time(start_time) * np.ones_like(x)

end_time = "1992-02-15"

t_bnds = np.array(
    [
        sd.utils.convert_time(start_time),
        sd.utils.convert_time(end_time),
    ]
)


@pytest.mark.parametrize("od,x,y,z,t", [("ecco", x, y, z, t)], indirect=["od"])
@pytest.mark.parametrize(
    "var_list",
    [
        ["ETAN", "maskC"],
        "SALT",
        ("UVELMASS", "VVELMASS"),
        {("UVELMASS", "VVELMASS"): (sd.KnW(), sd.KnW())},
    ],
)
def test_eulerian_oceinterp(od, var_list, x, y, z, t):
    ans = sd.OceInterp(od, var_list, x, y, z, t)
    assert isinstance(ans, list)


@pytest.mark.parametrize("ds", ["ecco"], indirect=True)
def test_xarray_interp(ds):
    ans = sd.OceInterp(ds, "ETAN", x, y, z, t, kernel_list=sd.KnW())
    assert isinstance(ans, list)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_diff_oceinterp(od):
    kernel_kwarg = {
        "hkernel": "dx",
        "h_order": 2,
        "inheritance": [[0, 1, 2, 3, 4, 5, 6, 7, 8]],
        "tkernel": "linear",
    }
    ans = sd.OceInterp(
        od, ("UVELMASS", "VVELMASS"), x, y, z, t, kernel_kwarg=kernel_kwarg
    )
    assert isinstance(ans[0], tuple)
    assert isinstance(ans[0][0], np.ndarray)


@pytest.mark.parametrize(
    "od,var_list,x,y,z,t,lagrangian,lagrange_kwarg",
    [
        (
            "ecco",
            ["SALT", "__particle.raw", "__particle.lat", "__particle.lon"],
            x,
            y,
            z,
            t_bnds,
            True,
            {"save_raw": True},
        )
    ],
    indirect=["od"],
)
@pytest.mark.parametrize("return_pt_time", [True, False])
@pytest.mark.filterwarnings("ignore::Warning")
def test_largangian_oceinterp(
    od, var_list, x, y, z, t, lagrangian, return_pt_time, lagrange_kwarg
):
    ans = sd.OceInterp(
        od,
        var_list,
        x,
        y,
        z,
        t,
        lagrangian=lagrangian,
        return_pt_time=return_pt_time,
        lagrange_kwarg=lagrange_kwarg,
    )
    if return_pt_time:
        ans = ans[1]
        assert ans[0] is not None
    assert isinstance(ans, list)
    for i, var in enumerate(var_list):
        assert isinstance(ans[i], list), f"{var} not in correct format"


@pytest.mark.parametrize(
    "var_list,x,y,z,t,lagrangian,error",
    [
        (
            ["__particle.lat", "__particle.lon"],
            x,
            y,
            z,
            t_bnds[:1],
            True,
            ValueError,
        ),
        (
            ["__particle.lat", "__particle.lon"],
            x,
            y,
            z,
            t,
            False,
            AttributeError,
        ),
        (None, x, y, z, t, False, ValueError),
        ([None], x, y, z, t, False, ValueError),
        ("ETAN", x, y, z, 0, False, ValueError),
    ],
)
@pytest.mark.filterwarnings("ignore:invalid value encountered in divide")
@pytest.mark.filterwarnings("ignore:divide by zero encountered in divide")
@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_oceinterp_error(od, var_list, x, y, z, t, lagrangian, error):
    with pytest.raises(error):
        sd.OceInterp(od, var_list, x, y, z, t, lagrangian=lagrangian)


@pytest.mark.parametrize("t", ["1992-02-01", np.datetime64("1992-02-01")])
@pytest.mark.parametrize("to_array", [True, False])
@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_flexible_time_format(od, t, to_array):
    if to_array:
        t = np.array([t for _ in x])
    res = sd.OceInterp(od, "ETAN", x, y, z, t)
    assert ~(np.isnan(res[0]).any())
    assert isinstance(res[0], np.ndarray)


@pytest.mark.parametrize(
    "od,var_list",
    [
        (
            "ecco",
            ["__particle.raw", "__particle.lat"],
        )
    ],
    indirect=["od"],
)
def test_lagrangian_warning(od, var_list):
    with pytest.warns(UserWarning):
        ans = sd.OceInterp(
            od,
            var_list,
            x,
            y,
            z,
            t,
            lagrangian=True,
            return_pt_time=False,
            return_in_between=True,
        )
        assert isinstance(ans, list)



import numpy as np
import pytest

import seaduck as sd
from seaduck import utils

# Set the number of particles here.
N = 9

# Increase this if you want more in x direction.
skew = 3

# Change the vertical depth of the particles here.
sqrtN = int(np.sqrt(N))

# Change the horizontal range here.
x = np.append(np.linspace(-180, 180, sqrtN * skew), -37.5)
y = np.append(np.linspace(-50, -70, sqrtN // skew), -56.73891)

x, y = np.meshgrid(x, y)
x = x.ravel()
y = y.ravel()
z = None
zz = np.ones_like(x) * (-10.0)

start_time = "1992-02-01"
t = utils.convert_time(start_time) * np.ones_like(x)
end_time = "1992-02-03"
tf = utils.convert_time(end_time)


@pytest.fixture
def p():
    od = sd.OceData(utils.get_dataset("aviso"))
    return sd.Particle(
        x=x,
        y=y,
        z=z,
        t=t,
        data=od,
        # save_raw = True,
        # transport = True,
        uname="u",
        vname="v",
        wname=None,
    )


@pytest.fixture
def ecco_p():
    od = sd.OceData(utils.get_dataset("ecco"))
    return sd.Particle(x=x, y=y, z=zz, t=t, data=od, transport=True)


@pytest.fixture
def kick_back_p():
    x = np.array([-38.594593, -37.512672, -36.42936, -34.08329, -35.06443])
    y = np.array([-77.95619, -77.97306, -77.98856, -77.25903, -76.86412])
    z = np.ones_like(x) * (-0.01)
    t = utils.convert_time(start_time) * np.ones_like(x)
    od = sd.OceData(utils.get_dataset("ecco"))
    return sd.Particle(x=x, y=y, z=z, t=t, data=od, free_surface="kick_back")


normal_stops = np.linspace(t[0], tf, 5)


def test_vol_mode(ecco_p):
    stops, raw = ecco_p.to_list_of_time(normal_stops=[t[0], tf])
    assert sd.get_masks.which_not_stuck(ecco_p).all()


def test_to_list_of_time(p):
    stops, raw = p.to_list_of_time(
        normal_stops=normal_stops, update_stops=[normal_stops[1]]
    )
    with pytest.warns(UserWarning):
        assert sd.get_masks.which_not_stuck(p).all()


def test_subset_update(p):
    np.random.seed(0)
    which = np.random.randint(1, size=p.N, dtype=bool)
    sub = p.subset(which)
    sub.lon += 1
    sub.lat += 1
    p.update_from_subset(sub, which)
    assert isinstance(sub, sd.Particle)
    assert np.allclose(p.ix[which], sub.ix)
    assert np.allclose(p.lon[which], sub.lon)


def test_subset_px_py(ecco_p):
    np.random.seed(1)
    which = np.random.randint(1, size=ecco_p.N, dtype=bool)
    ecco_p.px, ecco_p.py = ecco_p.get_px_py()
    sub = ecco_p.subset(which)
    sub.px, sub.py = sub.get_px_py()
    assert np.allclose(ecco_p.px[:, which], sub.px)
    assert np.allclose(ecco_p.py[:, which], sub.py)


@pytest.mark.parametrize("od", ["curv"], indirect=True)
def test_callback(od):
    curv_p = sd.Particle(
        y=np.array([70.5]),
        x=np.array([-14.0]),
        z=np.array([-10.0]),
        t=np.array([od.ts[0]]),
        data=od,
        uname="U",
        vname="V",
        wname="W",
        callback=lambda pt: pt.lon > -14.01,
    )
    curv_p.to_list_of_time(normal_stops=[od.ts[0], od.ts[-1]], update_stops=[])
    with pytest.warns(UserWarning):
        assert sd.get_masks.which_not_stuck(curv_p).all()


def test_note_taking_error(p):
    with pytest.raises(AttributeError):
        p.note_taking()


def test_no_time_midp_error(p):
    with pytest.raises(AttributeError):
        p.to_list_of_time(normal_stops=[0.0, 1.0])


def test_time_out_of_bound_error(ecco_p):
    with pytest.raises(ValueError):
        ecco_p.to_list_of_time(normal_stops=[0.0, 1.0], update_stops=[])


def test_multidim_uvw_array(ecco_p):
    ecco_p.it[0] += 1
    ecco_p.update_uvw_array()
    assert ecco_p.uarray.shape[0] == 2


def test_update_w_array(ecco_p):
    ecco_p.ocedata._ds["u0"] = ecco_p.ocedata["UVELMASS"].isel(time=0)
    ecco_p.ocedata._ds["v0"] = ecco_p.ocedata["VVELMASS"].isel(time=0)
    ecco_p.ocedata._ds["w0"] = ecco_p.ocedata["WVELMASS"].isel(time=0)
    delattr(ecco_p, "warray")
    ecco_p.uname = "u0"
    ecco_p.vname = "v0"
    ecco_p.wname = "w0"

    ecco_p.update_uvw_array()
    assert len(ecco_p.uarray.shape) == 4
    ecco_p.update_uvw_array()


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_wall_crossing(ecco_p, od):
    od["SN"] = np.array(od["SN"])
    od["CS"] = np.array(od["CS"])
    ecco_p.ocedata.readiness["h"] = "local_cartesian"

    ecco_p._cross_cell_wall_rel()
    assert (~np.isnan(ecco_p.ry)).any()


@pytest.mark.parametrize("od", ["curv"], indirect=True)
def test_wall_crossing_no_face(od):
    od._add_missing_cs_sn()
    od.readiness["h"] = "local_cartesian"
    curv_p = sd.Particle(
        y=np.array([70.5]),
        x=np.array([-14.0]),
        z=np.array([-10.0]),
        t=np.array([od.ts[0]]),
        data=od,
        uname="U",
        vname="V",
        wname="W",
        transport=True,
    )
    curv_p._cross_cell_wall_read()
    assert isinstance(curv_p.cs, np.ndarray)
    curv_p._cross_cell_wall_rel()
    assert (~np.isnan(curv_p.rx)).any()


@pytest.mark.parametrize("od", ["curv"], indirect=True)
def test_get_vol(od):
    curv_p = sd.Particle(
        y=np.array([70.5]),
        x=np.array([-14.0]),
        z=np.array([-10.0]),
        t=np.array([od.ts[0]]),
        data=od,
        uname="U",
        vname="V",
        wname="W",
        transport=True,
    )
    curv_p.get_vol()
    assert np.issubdtype(curv_p.vol.dtype, float)


def test_maxiteration(ecco_p):
    ecco_p.max_iteration = 1
    delattr(ecco_p, "px")
    with pytest.warns(UserWarning):
        ecco_p.to_next_stop(tf)


def test_abandon_ducks(ecco_p):
    N = len(ecco_p.izl_lin)
    ecco_p.izl_lin = (np.ones(N) * 50).astype(int)
    new_p = sd.get_masks.abandon_stuck(ecco_p)
    assert len(new_p.izl_lin) < N


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
@pytest.mark.parametrize("seed", list(range(5)))
def test_reproduce_latlon_oceanparcel(od, seed):
    np.random.seed(seed)
    x = np.random.uniform(-180, 180, 1)
    y = np.random.uniform(-90, 90, 1)
    t = sd.utils.convert_time("1992-02-01")
    z = -5.0
    rand_p = sd.Particle(x=x, y=y, z=z, t=t, data=od)
    rand_p._sync_latlondep_before_cross()
    assert np.allclose(rand_p.lon, x)
    assert np.allclose(rand_p.lat, y)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
@pytest.mark.parametrize("seed", [678, 15, 7])
def test_get_u_du_quant(seed, od):
    np.random.seed(seed)
    ind = tuple(np.random.randint([13, 89, 89]))
    face, iy, ix = ind
    # without boundary points
    x = od.XC[ind]
    y = od.YC[ind]
    t = tf
    z = np.random.uniform(0, -3000)
    rand_p = sd.Particle(x=x, y=y, z=z, t=t, data=od)
    u = rand_p.u * rand_p.dx
    du = rand_p.du * rand_p.dx
    v = rand_p.v * rand_p.dy
    dv = rand_p.dv * rand_p.dy
    w = rand_p.w * rand_p.dzl_lin
    dw = rand_p.dw * rand_p.dzl_lin

    it = rand_p.it[0]
    iz = rand_p.izl_lin[0] - 1
    u1 = float(od["UVELMASS"][it, iz, face, iy, ix])
    u2 = float(od["UVELMASS"][it, iz, face, iy, ix + 1])
    v1 = float(od["VVELMASS"][it, iz, face, iy, ix])
    v2 = float(od["VVELMASS"][it, iz, face, iy + 1, ix])
    w1 = float(od["WVELMASS"][it, iz + 1, face, iy, ix])
    w2 = float(od["WVELMASS"][it, iz, face, iy, ix])
    rx = rand_p.rx[0]
    ry = rand_p.ry[0]
    rz = rand_p.rzl_lin[0] - 0.5
    ushould = (0.5 + rx) * u2 + (0.5 - rx) * u1
    vshould = (0.5 + ry) * v2 + (0.5 - ry) * v1
    wshould = (0.5 + rz) * w2 + (0.5 - rz) * w1

    assert u1 != 0
    assert np.allclose(u2 - u1, du, atol=1e-18)
    assert np.allclose(v2 - v1, dv, atol=1e-18)
    assert np.allclose(w2 - w1, dw, atol=1e-18)
    assert np.allclose(ushould, u, atol=1e-18)
    assert np.allclose(vshould, v, atol=1e-18)
    assert np.allclose(wshould, w, atol=1e-18)


def test_kick_back(kick_back_p):
    tend = kick_back_p.analytical_step(-1e10)
    kick_back_p.cross_cell_wall(tend)
    assert np.isclose(kick_back_p.rzl_lin, 0.5).any()
